import os
import logging
from astropy.io import fits
import numpy as np

from .mask import Mask
from ..data import Spectrum


class MaskFromPath(Mask):
    """Reads a pre-calculated mask from another file of a given name.

    This class, when called, searches for a file of the given name in the directory specified in the configuration.
    If it exists, the extension of the given name (defaults to "GOODPIXELS") is read, converted into a Boolean array,
    and returned. If it doesn't exist, an empty mask is returned.
    """

    def __init__(self, path: str, fits_extension: str = 'GOODPIXELS', *args, **kwargs):
        """Initializes a new mask from a file in a given path.

        Args:
            path: Path to search in for file containing mask.
            fits_extension: FITS extension to read mask from.
        """
        Mask.__init__(self, *args, **kwargs)
        self._path = path
        self._fits_extension = fits_extension

    def __call__(self, spectrum: Spectrum, filename: str) -> np.ndarray:
        """Creates a new mask for a spectrum from a file of the same name in a given directory.

        Args:
            spectrum: Spectrum to create mask for.
            filename: Name of file containing spectrum to create mask for.

        Returns:
            Boolean array containing good pixel mask for given spectrum.
        """

        # build filename
        filename = os.path.join(self._path, os.path.basename(filename))

        # does it exist?
        if os.path.exists(filename):
            # read from file and return it as boolean array
            self.log.info('Reading additional mask from file %s...', filename)
            try:
                mask = fits.getdata(filename, self._fits_extension).astype(bool)
            except KeyError:
                self.log.warning('Could not find extension %s in FITS file, skipping.', self._fits_extension)
                return np.ones((len(spectrum.wave)), dtype=bool)

            # check length
            if len(spectrum) != len(mask):
                self.log.warning('Length of mask does not match length of spectum. Using empty mask.')
                return np.ones((len(spectrum.wave)), dtype=bool)

            # return mask
            return mask

        else:
            # return empty mask...
            self.log.warning('Could not find addition file for mask: %s', filename)
            return np.ones((len(spectrum.wave)), dtype=bool)


__all__ = ['MaskFromPath']



import argparse
import importlib
import os
import pkgutil
import spexxy.tools
import logging


def main():
    # init logging
    logging.basicConfig(format='[%(asctime)s] %(message)s', level=logging.INFO)
    logging.captureWarnings(True)

    # init parser
    parser = argparse.ArgumentParser(description='spexxy command line interface',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(help='sub-command help')

    # list modules in spexxy.methods
    pkgpath = os.path.dirname(spexxy.tools.__file__)
    modules = [name for _, name, _ in pkgutil.iter_modules([pkgpath])]

    # loop modules
    for m in modules:
        # import module
        mod = importlib.import_module('spexxy.tools.' + m)
        # add subparser
        try:
            mod.add_parser(subparsers)
        except AttributeError:
            continue

    # parse arguments
    args = parser.parse_args()

    # and call method
    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == '__main__':
    main()





import os

import numpy as np
import pandas as pd
from typing import List, Union

from .weight import Weight
from ..data import Spectrum


class WeightFromGridNearest(Weight):
    """
    This class loads the weights from a grid depending on the initial values of the fit parameters by choosing the
     nearest neighbour in the grid. It returns an array containing the weights.
    """

    def __init__(self, filename, initial: float = 0., max_line_depth: float = 0.5, center_weight: float = 1.,
                 max_step: int = 1, mask_lines: Union[bool, str, List] = True, max_change=(300, 0.3), *args, **kwargs):
        """
        Initializes a new weight.

        Args:
            filename: Name of grid file.
            initial: Initial value for the whole weight array.
            max_line_depth: Central pixel for lines with larger line depth are masked out.
            center_weight: Factor that increases the weight of the central pixel of each line.
            max_step: In iteration steps <= max_step new weights are loaded from the grid.
            mask_lines: List of absorption lines that are always masked out in their centers.
        """

        Weight.__init__(self, *args, **kwargs)

        if 'objects' in kwargs:
            self.objects = kwargs['objects']

        self._initial = initial
        self._max_line_depth = max_line_depth
        self._center_weight = center_weight
        self._max_step = max_step
        self._max_change = sorted(max_change, reverse=True)

        if mask_lines:
            if isinstance(mask_lines, bool):
                self._mask_lines = 'default'
            elif isinstance(mask_lines, list):
                self._mask_lines = []
                for line in mask_lines:
                    if len(line) == 2:
                        self._mask_lines.append(line + [-0.5, 6.5])
                    else:
                        self._mask_lines.append(line)
            elif isinstance(mask_lines, str):
                df = pd.read_csv(os.path.expandvars(mask_lines))
                df.loc[df['logg_min'].isna(), 'logg_min'] = -0.5
                df.loc[df['logg_max'].isna(), 'logg_max'] = 6.5

                self._mask_lines = df.to_numpy()
        else:
            self._mask_lines = mask_lines

        # expand filename
        filename = os.path.expandvars(filename)

        # get grid's root path
        self._root = os.path.dirname(filename)

        # load CSV
        self._data = pd.read_csv(filename, index_col=False)

        # get all parameters, by removing 'Filename' from list of columns
        self._parameters = list(self._data.columns)
        self._parameters.remove('Filename')

        # we assume that all parameters are floats, so treat them as such
        for name in self._parameters:
            self._data[name] = self._data[name].apply(lambda x: float(x))

        # get grid axes
        self._axes = [np.array(sorted(self._data[p].unique())) for p in self._parameters]

        # remove axes that contain only a single value
        for i, p in enumerate(self._parameters):
            if len(self._axes[i]) <= 1:
                del self._axes[i]
                self._parameters.remove(p)

        self._data.set_index(keys=self._parameters, inplace=True)

        # initialize step counter
        self._step = 1

        # values of the fit parameter from previous iteration step
        self._previous_values = None

        self._filename = None

        # weights will be stored for next iterations
        self._weights = None
        self._neighbour = None

        # save initial parameters
        self._initial_values = None

        self._logg = None

    def __call__(self, spectrum: Spectrum, filename: str) -> np.ndarray:
        """
        Creates and returns weight array.

        Args:
            spectrum: Spectrum to create weight for.
            filename: Name of spectrum file.

        Returns:
             Array containing the weight for given spectrum.
        """

        # save initial values
        if self._initial_values is None:
            self._initial_values = {}
            for cmp in self.objects['init_iter'].values():
                for param_name in cmp.param_names:
                    self._initial_values[param_name] = cmp[param_name]

                    if param_name == 'logg' and self._logg is None:
                        self._logg = cmp[param_name]

                break

        # load new weights if the fit parameters changed significantly
        new_weights = False
        if self._previous_values is not None:
            for param in self._parameters:
                if new_weights:
                    break
                for cmp in self.objects['init_iter'].values():
                    for param_name in cmp.param_names:
                        if param.lower() != param_name.lower():
                            continue

                        if param.lower() == 'teff':
                            # did Teff change by more than 300K?
                            new_weights = abs(
                                self._previous_values[self._parameters.index(param)] - cmp[param_name]) > self._max_change[0]
                        else:
                            # did FeH, Alpha or logg change by more than 0.3 dex?
                            new_weights = abs(
                                self._previous_values[self._parameters.index(param)] - cmp[param_name]) > self._max_change[1]

        # are current parameter values identical with initial values?
        if self._step > 1:
            tmp = []
            for cmp in self.objects['init_iter'].values():
                for param_name in cmp.param_names:
                    tmp.append(cmp[param_name] == self._initial_values[param_name])

                break

            # component is reset to initial values if the fit restarts with a damping factor, in that case the iteration
            #  step needs to be reset as well
            if np.all(tmp):
                self._step = 1

        # load new weights if max_step has not been reached or fit parameters changed significantly
        if (self._step <= self._max_step) or new_weights:
            if new_weights:
                self._step = 1

            # get parameters from component
            params = []
            for param in self._parameters:
                for cmp in self.objects['init_iter'].values():
                    for param_name in cmp.param_names:
                        if param.lower() != param_name.lower():
                            continue

                        params.append(cmp[param_name])

                    break

            # save current parameters for next step
            self._previous_values = params.copy()

            # find nearest neighbour
            self._neighbour = []
            for i, p in enumerate(params):
                self._neighbour.append(self._axes[i][np.argmin(np.abs(self._axes[i] - p))])

            # save filename of weight table
            self._filename = self._data.loc[tuple(self._neighbour)].Filename

        # load weights
        w = self._load_weights(spectrum)

        # increase step counter
        self._step += 1

        return w

    def _load_weights(self, spectrum: Spectrum):
        """
        Load CSV file from grid and create weight array.

        Args:
            filename: Filename of CSV file that contains the weights.
            spectrum: Spectrum to create the weight for.

        Returns:
             Weight array.
        """

        # load table containing the weights
        df = pd.read_csv(os.path.join(self._root, self._filename))

        # consider only weights for iteration steps lower/equal than the given one
        df = df[df['step'] <= self._step]

        # initialize weight array
        weights = np.zeros(spectrum.wave.shape) + self._initial

        # write weights to array
        for i, row in df.iterrows():
            if isinstance(self._mask_lines, list) or isinstance(self._mask_lines, np.ndarray):
                if self._mask_centers(row, self._mask_lines, weights, spectrum):
                    continue
            elif self._mask_lines == 'default':
                if self._mask_default_lines(row, weights, spectrum):
                    continue

            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            # if line depth larger than given threshold mask out the central region otherwise increase weight of
            # central pixel by given factor
            if row['line_depth'] > self._max_line_depth:
                # if region spans more than 10 wavelength pixel mask out the 3 central pixel otherwise only the central
                # one
                if (row['wave_end'] - row['wave_start']) >= 12:
                    i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
                    weights[i-1:i+2] = 0
                else:
                    weights[np.argmin(np.abs(spectrum.wave - row['wave_center']))] = 0
            else:
                weights[np.argmin(np.abs(spectrum.wave - row['wave_center']))] *= self._center_weight

        return weights

    def _mask_default_lines(self, row: pd.Series, weights: np.ndarray, spectrum: Spectrum):
        # Halpha
        if (row['wave_center'] < 6566.) & (row['wave_center'] > 6557.) & (self._logg <= 3.5):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 1:i + 2] = 0
            return True
        elif (row['wave_center'] < 6566.) & (row['wave_center'] > 6557.) & (self._logg > 3.5):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i] = 0
            return True

        # Hbeta
        if (row['wave_center'] < 4867.) & (row['wave_center'] > 4857.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i] = 0
            return True

        # FeI line
        if (row['wave_center'] < 5272.) and (row['wave_center'] > 5267.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 1:i + 2] = 0
            return True

        # Ca triplet
        if (row['wave_center'] < 8508.) and (row['wave_center'] > 8490.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 2:i + 3] = 0
            return True

        if (row['wave_center'] < 8553.) and (row['wave_center'] > 8530.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 2:i + 3] = 0
            return True

        if (row['wave_center'] < 8672.) and (row['wave_center'] > 8651.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 2:i + 3] = 0
            return True

        return False

    def _mask_centers(self, row: pd.Series, lines: Union[list, np.ndarray], weights: np.ndarray, spectrum: Spectrum):
        for center, npix, logg_min, logg_max in lines:
            if (row['wave_start'] < center) and (row['wave_end'] > center) and (self._logg < logg_max) and (
                    self._logg >= logg_min):
                weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

                i = np.argmin(np.abs(spectrum.wave - row['wave_center']))

                if npix % 2 == 0:
                    weights[int(i-npix//2):int(i+npix//2)] = 0
                else:
                    weights[int(i-npix//2):int(i+npix//2+1)] = 0

                return True

        return False


__all__ = ['WeightFromGridNearest']



from .spectrumcomponent import SpectrumComponent
from ..interpolator import Interpolator


class GridComponent(SpectrumComponent):
    """A Grid component takes an interpolator and adds LOSVD parameters."""

    def __init__(self, interpolator: Interpolator, name: str = "STAR", *args, **kwargs):
        """Initializes a new Grid component.

        Parameters
        ----------
        interpolator : Interpolator
            The interpolator to use for the component
        name : str
            Name of the component
        """
        SpectrumComponent.__init__(self, name, *args, **kwargs)
        self.log.info('Initializing grid component "%s"...', name)

        # get interpolator
        self._interpolator = self.get_objects(interpolator, Interpolator, 'interpolators')
        if isinstance(self._interpolator, list):
            self._interpolator = self._interpolator[0]

        # add parameters of interpolator
        for a in self._interpolator.axes():
            self.log.info('Found parameter %s with min=%.2f, max=%.2f, and initial=%.2f.',
                          a.name, a.min, a.max, a.initial)
            self.set(a.name, min=a.min, max=a.max, value=a.initial)

    def _model_func(self):
        """Get spectrum with given parameters.

        Returns
        -------
        Any
            Result from interpolator
        """

        # get values as tuple
        values = tuple([self[a.name] for a in self._interpolator.axes()])

        # interpolate
        return self._interpolator(values)


__all__ = ['GridComponent']



import os
import astropy.io.fits as pyfits
import numpy as np
import scipy.linalg
from typing import List, Tuple

from . import Interpolator
from ..data import Spectrum
from ..grid import GridAxis


class UlyssInterpolator(Interpolator):
    """Interpolator that works with Ulyss input files."""

    def __init__(self, filename: str, *args, **kwargs):
        """Initializes a new Ulyss interpolator.

        Args:
            filename: Name of file containing ulyss interpolator.
        """
        Interpolator.__init__(self, *args, **kwargs)

        # store it
        self.filename = os.path.expandvars(filename)

        # get info
        hdr = pyfits.getheader(self.filename, 0)
        self.wStart = hdr["CRVAL1"]
        self.wStep = hdr["CDELT1"]
        self.wCount = hdr["NAXIS1"]
        self.waveMode = Spectrum.Mode.LAMBDA
        if hdr["CTYPE1"] == "AWAV-LOG":
            self.waveMode = Spectrum.Mode.LOGLAMBDA

        # load coefficients
        self._coeffsWarm = pyfits.getdata(self.filename, 0)[:23]
        self._coeffsHot = pyfits.getdata(self.filename, 1)[:19]
        self._coeffsCold = pyfits.getdata(self.filename, 2)[:23]

    def axes(self) -> List[GridAxis]:
        """Returns information about the axes.

        Returns:
            List of GridAxis objects describing the grid's axes.
        """
        return [
            GridAxis(name='Teff', initial=6000., min=2300., max=15000.),
            GridAxis(name='logg', initial=4.5, min=0., max=6.),
            GridAxis(name='FeH', initial=-1.5, min=-4., max=1.)
        ]

    @staticmethod
    def _poly_matrix(params: Tuple) -> np.ndarray:
        """Create the matrix used in the polynomial evaluation for the given params.

        Args:
            params: Parameters to interpolate at

        Returns:
            Matrix that can be used for the actual interpolation
        """
        # primary scaling
        Teff = np.log10(params[0]) - 3.7617
        logg = params[1] - 4.44
        FeH = params[2]

        # matrix
        mat = np.zeros((3, 23))
        for i in [0, 1, 2]:
            # secondary scaling
            teffc = (Teff + 0.1) if i == 2 else Teff
            tt = teffc / 0.2
            tt2 = tt * tt - 1.

            # set it
            mat[i, 0] = 1
            mat[i, 1] = tt
            mat[i, 2] = FeH
            mat[i, 3] = logg
            mat[i, 4] = tt * tt
            mat[i, 5] = tt * tt2
            mat[i, 6] = tt2 * tt2
            mat[i, 7] = tt * FeH
            mat[i, 8] = tt * logg
            mat[i, 9] = tt2 * logg
            mat[i, 10] = tt2 * FeH
            mat[i, 11] = logg * logg
            mat[i, 12] = FeH * FeH
            mat[i, 13] = tt * tt2 * tt2
            mat[i, 14] = tt * logg * logg
            mat[i, 15] = logg * logg * logg
            mat[i, 16] = FeH * FeH * FeH
            mat[i, 17] = tt * FeH * FeH
            mat[i, 18] = logg * FeH
            mat[i, 19] = logg * logg * FeH
            mat[i, 20] = logg * FeH * FeH
            mat[i, 21] = np.exp(tt) - 1. - tt * (1. + tt / 2. +
                                                 tt * tt / 6. + tt * tt * tt / 24. + tt * tt * tt * tt / 120.)
            mat[i, 22] = np.exp(tt * 2.) - 1. - 2. * tt * (1. + tt +
                                                           2. / 3. * tt * tt + tt * tt * tt / 3. +
                                                           tt * tt * tt * tt * 2. / 15.)

        # return matrix
        return mat

    def __call__(self, params: Tuple) -> Spectrum:
        """Interpolates at the given parameter set.

        Args:
            params: Parameter set to interpolate at.

        Returns:
            Interpolated spectrum at given position.
        """

        # parameters
        teff = np.log10(params[0]) - 3.7617
        grav = params[1] - 4.44

        # create matrix
        mat = UlyssInterpolator._poly_matrix(params)

        # read coefficients
        if teff <= np.log10(9000.) - 3.7617:
            t1 = np.dot(mat[0, :], self._coeffsWarm)
        if teff >= np.log10(7000.) - 3.7617:
            t2 = np.dot(mat[1, :19], self._coeffsHot)
        if teff <= np.log10(4550.) - 3.7617:
            t3 = np.dot(mat[2, :], self._coeffsCold)

        # interpolate spectrum
        flux = None
        if teff <= np.log10(7000.) - 3.7617:
            if teff > np.log10(4550.) - 3.7617:
                flux = t1
            elif teff > np.log10(4000.) - 3.7617:
                q = (teff - np.log10(4000.) + 3.7617) / (np.log10(4550.) -
                                                         np.log10(4000.))
                flux = q * t1 + (1. - q) * t3
            else:
                flux = t3
        elif teff >= np.log10(9000.) - 3.7617:
            flux = t2
        else:
            q = (teff - np.log10(7000.) + 3.7617) / (np.log10(9000.) -
                                                     np.log10(7000.))
            flux = q * t2 + (1. - q) * t1

        # create spectrum and return it
        return Spectrum(flux=flux, wave_start=self.wStart, wave_step=self.wStep, wave_mode=self.waveMode)

    @staticmethod
    def create(files: List[str], output: str):
        """Create a new ulyss interpolator from a set of spectrum files.

        Args:
            files: List of spectrum files.
            output: Output file name.
        """

        # prepare lists for data
        teffs = [[], [], []]
        loggs = [[], [], []]
        fehs = [[], [], []]
        fluxes = [[], [], []]

        # load all spectra
        for i, f in enumerate(files):
            print("({0:d}/{1:d}) {2:s}".format(i, len(files), f))
            # open file
            fits = pyfits.open(f, memmap=False)

            # get params
            hdr = fits[0].header
            teff = hdr["PHXTEFF"]
            logg = hdr["PHXLOGG"]
            feh = hdr["PHXM_H"]

            # spectrum flux
            flux = fits[0].data
            flux /= np.mean(flux)

            # store it
            if teff < 4550:
                teffs[2].append(teff)
                loggs[2].append(logg)
                fehs[2].append(feh)
                fluxes[2].append(flux)
            if 4000 < teff < 9000:
                teffs[0].append(teff)
                loggs[0].append(logg)
                fehs[0].append(feh)
                fluxes[0].append(flux)
            if teff > 7000:
                teffs[1].append(teff)
                loggs[1].append(logg)
                fehs[1].append(feh)
                fluxes[1].append(flux)

            # close file
            fits.close()

        # our output HDUs
        hdus = []

        # loop regimes
        for regime in [0, 1, 2]:
            # create matrix with polynomial evaluations
            num = len(fluxes[regime])
            c = 19 if regime == 1 else 23
            polys = np.zeros((num, c))
            for i in range(num):
                # eval poly
                p = UlyssInterpolator._poly_matrix([teffs[regime][i], loggs[regime][i], fehs[regime][i]])
                # store it
                polys[i, :c] = p[regime, :c]

            # now loop wavelength bins
            coeffs = np.zeros((c, len(fluxes[regime][0])))
            for w in range(len(fluxes[regime][0])):
                # get all fluxes for this wavelength bin
                flx = [fluxes[regime][i][w] for i in range(len(fluxes[regime]))]
                # fit
                coeffs[:, w] = scipy.linalg.lstsq(polys, flx)[0]
                #print regime, w, coeffs[:, w]

            # store as ImageHDU
            if regime == 0:
                hdu = pyfits.PrimaryHDU(coeffs.astype(np.float32))
                hdr = pyfits.getheader(files[0])
                hdu.header["CRPIX1"] = 1
                hdu.header["CRVAL1"] = hdr["CRVAL1"]
                hdu.header["CDELT1"] = hdr["CDELT1"]
                hdu.header["CD1_1"] = hdr["CDELT1"]
                hdu.header["CTYPE1"] = hdr["CTYPE1"]
            else:
                hdu = pyfits.ImageHDU(coeffs.astype(np.float32))
                name = {1: "hot", 2: "cold"}
                hdu.header["EXTNAME"] = name[regime]
            hdus.append(hdu)

        # create FITS file
        hdulist = pyfits.HDUList(hdus)
        hdulist.writeto(output)


__all__ = ['UlyssInterpolator']



import logging
import os
from typing import Callable
import numpy as np
import pandas as pd
import argparse

from spexxy.data import SpectrumFits
from spexxy.interpolator import SplineInterpolator

from spexxy.grid import FilesGrid, Grid

log = logging.getLogger(__name__)


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('fillholes', help='Fill holes in a file grid',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('grid', type=str, help='Name of grid CSV', default='grid.csv')
    parser.add_argument('holes', type=str, help='Name of holes CSV', default='holes.csv')
    parser.add_argument('output', type=str, help='Directory to store interpolated spectra in', default='interpolated/')

    # argparse wrapper for fill_holes
    def run(args):
        grid = FilesGrid(args.grid)
        holes = pd.read_csv(args.holes, index_col=False)
        fill_holes(grid, holes, args.output)
    parser.set_defaults(func=run)


def fill_holes(grid: Grid, holes: Grid, output: str):
    # create spline interpolator on grid
    ip = SplineInterpolator(grid, n=100)

    # loop holes
    for i, row in holes.iterrows():
        # build params
        params = tuple([row[key] for key in grid.axis_names()])
        log.info('(%d/%d) %s', i, len(holes), params)

        # interpolate
        try:
            spec = ip(params)
        except KeyError:
            log.warning('  Could not interpolate spectrum.')
            continue

        # create filename
        filename = os.path.join(output, '_'.join(['%.1f' % f for f in params]) + '.fits')

        # save it
        SpectrumFits(spec=spec).save(filename)



import os

import numpy as np
import pandas as pd
from typing import List, Union, Tuple

from .weight import Weight
from .fromgridnearest import WeightFromGridNearest
from ..data import Spectrum


class WeightFromGrid(Weight):
    """
    This class loads the weights from a grid depending on the initial values of the fit parameters by linear
    interpolation. It returns an array containing the weights.
    """

    def __init__(self, filename, initial: float = 0., max_line_depth: float = 0.5, center_weight: float = 1.,
                 max_step: int = 1, mask_lines: Union[bool, str, List] = True, max_change=(300, 0.3), *args, **kwargs):
        """
        Initializes a new weight.

        Args:
            filename: Name of grid file.
            initial: Initial value for the whole weight array.
            max_line_depth: Central pixel for lines with larger line depth are masked out.
            center_weight: Factor that increases the weight of the central pixel of each line.
            max_step: In iteration steps <= max_step new weights are loaded from the grid.
            mask_lines: List of absorption lines that are always masked out in their centers.
        """

        Weight.__init__(self, *args, **kwargs)

        # expand filename
        filename = os.path.expandvars(filename)

        self.filename = filename

        self._initial = initial
        self._max_line_depth = max_line_depth
        self._center_weight = center_weight
        self._max_step = max_step
        self._max_change = sorted(max_change, reverse=True)

        if mask_lines:
            if isinstance(mask_lines, bool):
                self._mask_lines = 'default'
            elif isinstance(mask_lines, list):
                self._mask_lines = []
                for line in mask_lines:
                    if len(line) == 2:
                        self._mask_lines.append(line + [-0.5, 6.5])
                    else:
                        self._mask_lines.append(line)
            elif isinstance(mask_lines, str):
                df = pd.read_csv(os.path.expandvars(mask_lines))
                df.loc[df['logg_min'].isna(), 'logg_min'] = -0.5
                df.loc[df['logg_max'].isna(), 'logg_max'] = 6.5

                self._mask_lines = df.to_numpy()
        else:
            self._mask_lines = mask_lines

        # get grid's root path
        self._root = os.path.dirname(filename)

        # load CSV
        self._data = pd.read_csv(filename, index_col=False)

        # get all parameters, by removing 'Filename' from list of columns
        self._parameters = list(self._data.columns)
        self._parameters.remove('Filename')

        # we assume that all parameters are floats, so treat them as such
        for name in self._parameters:
            self._data[name] = self._data[name].apply(lambda x: float(x))

        # get grid axes
        self._axes = [sorted(self._data[p].unique()) for p in self._parameters]

        # remove axes that contain only a single value
        for i, p in enumerate(self._parameters):
            if len(self._axes[i]) <= 1:
                del self._axes[i]
                self._parameters.remove(p)

        # initialize step counter
        self._step = 1

        # values of the fit parameter from previous iteration step
        self._previous_values = None

        # save weight array
        self._weights = None

        # save initial parameters
        self._initial_values = None

        self._logg = None

    def __call__(self, spectrum: Spectrum, filename: str) -> np.ndarray:
        """
        Creates and returns weight array.

        Args:
            spectrum: Spectrum to create weight for.
            filename: Name of spectrum file.

        Returns:
             Array containing the weight for given spectrum.
        """

        # save initial values
        if self._initial_values is None:
            self._initial_values = {}
            for cmp in self.objects['init_iter'].values():
                for param_name in cmp.param_names:
                    self._initial_values[param_name] = cmp[param_name]

                    if param_name == 'logg' and self._logg is None:
                        self._logg = cmp[param_name]

                break

        # load new weights if the fit parameters changed significantly
        new_weights = False
        if self._previous_values is not None:
            for param in self._parameters:
                if new_weights:
                    break
                for cmp in self.objects['init_iter'].values():
                    for param_name in cmp.param_names:
                        if param.lower() != param_name.lower():
                            continue

                        if param.lower() == 'teff':
                            # did Teff change significantly?
                            new_weights = abs(
                                self._previous_values[self._parameters.index(param)] - cmp[param_name]) > self._max_change[0]
                        else:
                            # did FeH, Alpha or logg change significantly?
                            new_weights = abs(
                                self._previous_values[self._parameters.index(param)] - cmp[param_name]) > self._max_change[1]

        # are current parameter values identical with initial values?
        if self._step > 1:
            tmp = []
            for cmp in self.objects['init_iter'].values():
                for param_name in cmp.param_names:
                    tmp.append(cmp[param_name] == self._initial_values[param_name])

                break

            # component is reset to initial values if the fit restarts with a damping factor, in that case the iteration
            #  step needs to be reset as well
            if np.all(tmp):
                self._step = 1

        # load new weights if max_step has not been reached or fit parameters changed significantly
        if (self._step <= self._max_step) or new_weights:
            if new_weights:
                self._step = 1

            # get parameters from component
            params = []
            for param in self._parameters:
                for cmp in self.objects['init_iter'].values():
                    for param_name in cmp.param_names:
                        if param.lower() != param_name.lower():
                            continue

                        params.append(cmp[param_name])

                    break

            # save current parameters for next step
            self._previous_values = params.copy()

            # interpolate weight for given values, use nearest neighbour if values are outside of the grid
            try:
                self._weight_table = self._interpolate(tuple(params))
            except KeyError:
                self._weight_table = None

                if self._mask_lines == 'default':
                    w = WeightFromGridNearest(self.filename, self._initial, self._max_line_depth, self._center_weight,
                                              self._max_step, objects=self.objects)
                else:
                    w = WeightFromGridNearest(self.filename, self._initial, self._max_line_depth, self._center_weight,
                                              self._max_step, self._mask_lines, objects=self.objects)

                self._weights = {step: w(spectrum, filename) for step in range(1, 6)}

        if self._weight_table is None:
            if self._step <= 5:
                w = self._weights[self._step]
                self._step += 1

                return w

            return self._weights[5]

        w = self._get_weight_array(self._weight_table, spectrum)
        self._step += 1

        return w

    def _interpolate(self, params: Tuple, axis: int = None) -> pd.DataFrame:
        # no axis given, start at latest
        if axis is None:
            axis = len(self._axes) - 1

        if params[axis] < min(self._axes[axis]) or params[axis] > max(self._axes[axis]):
            raise KeyError('Requested parameters are outside the grid.')

        # let's get all possible values for the given axis
        axisValues = self._axes[axis].copy()

        # if params[axis] is on axis; return it directly
        if params[axis] in axisValues:
            if axis == 0:
                return self._load_weight_table(params)
            else:
                return self._interpolate(tuple(params), axis - 1)

        # find the next lower and the next higher axis value
        p_lower = self._neighbour(tuple(params), axis, 0)
        p_higher = self._neighbour(tuple(params), axis, 1)

        if p_lower is None or p_higher is None:
            raise KeyError('No direct neighbours found in grid.')

        # get axis values
        x_lower = p_lower[axis]
        x_higher = p_higher[axis]

        # get data for p_lower and p_higher
        if axis == 0:
            lower_data = self._load_weight_table(p_lower)
            higher_data = self._load_weight_table(p_higher)
        else:
            lower_data = self._interpolate(p_lower)
            higher_data = self._interpolate(p_higher)

        # interpolate
        f = (params[axis] - x_lower) / (x_higher - x_lower)

        # add interpolation weight to table
        lower_data['w'] = 1. - f
        higher_data['w'] = f

        df = pd.concat([lower_data, higher_data])
        df = df.sort_values(by=['wave_center'])
        cols = list(df.columns)
        cols.remove('wave_center')

        # assign identical values to wave centers less than 0.3 Angstrom apart
        centers = df['wave_center'].values
        while True:
            delta = np.ediff1d(centers)
            mask = (np.abs(delta) <= 0.3) & (delta != 0.)
            centers[1:][mask] = centers[:-1][mask]

            if not np.any(mask):
                break

        # average lines with identical wave centers together
        df['wave_center'] = centers
        df_grouped = df.groupby('wave_center')
        df = df_grouped.filter(lambda x: len(x) == 1).copy()
        df['weights'] *= df['w'].values

        # average lines showing up in both tables
        df2 = df_grouped.filter(lambda x: (len(x) == 2) & (x['w'].sum() == 1)).copy()
        w = df2['w'].values[:, None]
        df2.loc[:, cols] *= w
        df2 = df2.groupby('wave_center').sum()
        df2 = df2.reset_index()
        cols.insert(1, 'wave_center')
        df2 = df2.loc[:, cols]

        df = pd.concat([df, df2]).sort_values('wave_center')
        df = df.drop(columns=['w'])

        df['step'] = df['step'].apply(lambda x: np.around(x, decimals=0))

        return df

    def _load_weight_table(self, params: Tuple) -> pd.DataFrame:
        """
        Load CSV file containing the weights for the given set of parameters.
        """

        grid = self._data.copy()
        for i, p in enumerate(self._parameters):
            grid = grid[grid[p] == params[i]]

        filename = grid.Filename.values[0]

        return pd.read_csv(os.path.join(self._root, filename))

    def _get_weight_array(self, df: pd.DataFrame, spectrum: Spectrum):
        """
        Create weight array from the given table for the given spectrum.

        Args:
            df: Table containing the weights for each absorption line considered in that fitting mask.
            spectrum: Spectrum for which the weight array is created.

        Returns:
             Weight array.
        """

        # consider only weights for iteration steps lower/equal than the given one
        df = df[df['step'] <= self._step]
        df = df[df['weights'] > 10.]

        # initialize weight array
        weights = np.zeros(spectrum.wave.shape) + self._initial

        # write weights to array
        for i, row in df.iterrows():
            if isinstance(self._mask_lines, list) or isinstance(self._mask_lines, np.ndarray):
                if self._mask_centers(row, self._mask_lines, weights, spectrum):
                    continue
            elif self._mask_lines == 'default':
                if self._mask_default_lines(row, weights, spectrum):
                    continue

            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            # if line depth larger than given threshold mask out the central region otherwise increase weight of
            # central pixel by given factor
            if row['line_depth'] > self._max_line_depth:
                # if region spans more than 10 wavelength pixel mask out the 3 central pixel otherwise only the central
                # one
                if (row['wave_end'] - row['wave_start']) >= 12:
                    i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
                    weights[i-1:i+2] = 0
                else:
                    weights[np.argmin(np.abs(spectrum.wave - row['wave_center']))] = 0
            else:
                weights[np.argmin(np.abs(spectrum.wave - row['wave_center']))] *= self._center_weight

        return weights

    def _neighbour(self, params: Tuple, axis: int, distance: int = 1):
        """Finds a neighbour on the given axis for the given value in the given distance.

        Args:
            params: Paremeter tuple to search neighbour from.
            axis: Axis to search for
            distance: Distance in which to find neighbour.
                >0:  Find larger neighbours, i.e. 0 next larger value, 1 the one after that, etc
                <=0:  Find smaller neighbouars, i.e. 0 next smaller value (or value itself), -1 the before that, etc

        Returns:
            New parameter tuple with neighbour on the given axis.
        """

        # find neighbour in axis
        values = self._axes[axis]
        value = None
        # loop all values
        for i in range(len(values)):
            # found value?
            if values[i] <= params[axis] < values[i + 1]:
                # index of neighbour
                ii = i + distance
                # does it exist?
                if 0 <= ii < len(values):
                    value = values[ii]

        if value is None:
            return None

        # create new tuple
        p = list(params)
        p[axis] = value

        return tuple(p)

    def _mask_default_lines(self, row: pd.Series, weights: np.ndarray, spectrum: Spectrum):
        # Halpha
        if (row['wave_center'] < 6566.) & (row['wave_center'] > 6557.) & (self._logg <= 3.5):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 1:i + 2] = 0
            return True
        elif (row['wave_center'] < 6566.) & (row['wave_center'] > 6557.) & (self._logg > 3.5):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i] = 0
            return True

        # Hbeta
        if (row['wave_center'] < 4867.) & (row['wave_center'] > 4857.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i] = 0
            return True

        # FeI line
        if (row['wave_center'] < 5272.) and (row['wave_center'] > 5267.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 1:i + 2] = 0
            return True

        # Ca triplet
        if (row['wave_center'] < 8508.) and (row['wave_center'] > 8490.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 2:i + 3] = 0
            return True

        if (row['wave_center'] < 8553.) and (row['wave_center'] > 8530.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 2:i + 3] = 0
            return True

        if (row['wave_center'] < 8672.) and (row['wave_center'] > 8651.):
            weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

            i = np.argmin(np.abs(spectrum.wave - row['wave_center']))
            weights[i - 2:i + 3] = 0
            return True

        return False

    def _mask_centers(self, row: pd.Series, lines: Union[list, np.ndarray], weights: np.ndarray, spectrum: Spectrum):
        for center, npix, logg_min, logg_max in lines:
            if (row['wave_start'] < center) and (row['wave_end'] > center) and (self._logg < logg_max) and (
                    self._logg >= logg_min):
                weights[(spectrum.wave >= row['wave_start']) & (spectrum.wave <= row['wave_end'])] += row['weights']

                i = np.argmin(np.abs(spectrum.wave - row['wave_center']))

                if npix % 2 == 0:
                    weights[int(i-npix//2):int(i+npix//2)] = 0
                else:
                    weights[int(i-npix//2):int(i+npix//2+1)] = 0

                return True

        return False


__all__ = ['WeightFromGrid']



import numpy as np

from .mask import Mask
from ..data import Spectrum


class MaskNegative(Mask):
    """Masks negative fluxes in a spectrum.

    This class, when called, creates a mask that masks all negative pixels in the given spectrum.
    """

    def __init__(self, *args, **kwargs):
        """Initializes a new mask for masking negative pixels in a spectrum."""
        Mask.__init__(self, *args, **kwargs)

    def __call__(self, spectrum: Spectrum, filename: str = None) -> np.ndarray:
        """Creates a new mask for the given spectrum masking all negative pixels.

        Args:
            spectrum: Spectrum to create mask for.
            filename: Name of file containing spectrum to create mask for (unused).

        Returns;
            Boolean array containing good pixel mask for given spectrum.
        """

        # mask all negative pixels, i.e. create a mask with only positive fluxes, and return it
        self.log.info('Masking all negative pixels in spectrum...')
        return spectrum.flux > 0


__all__ = ['MaskNegative']



import copy
import shutil
from tempfile import mkdtemp
from typing import List, Any, Tuple, Dict, Union
import pandas as pd
import os

from .grid import Grid, GridAxis
from ..data import Spectrum


ABUND_AGSS = [
    (1, 'H', 12., 2),
    (2, 'He', 10.93, 2),
    (3, 'Li', 3.26, 1),
    (4, 'Be', 1.38, 1),
    (5, 'B', 2.79, 1),
    (6, 'C', 8.43, 2),
    (7, 'N', 7.83, 2),
    (8, 'O', 8.69, 2),
    (9, 'F', 4.56, 1),
    (10, 'Ne', 7.93, 2),
    (11, 'Na', 6.24, 1),
    (12, 'Mg', 7.60, 2),
    (13, 'Al', 6.45, 2),
    (14, 'Si', 7.51, 2),
    (15, 'P', 5.41, 1),
    (16, 'S', 7.12, 2),
    (17, 'Cl', 5.50, 1),
    (18, 'Ar', 6.40, 1),
    (19, 'K', 5.08, 1),
    (20, 'Ca', 6.34, 1),
    (21, 'Sc', 3.15, 1),
    (22, 'Ti', 4.95, 1),
    (23, 'V', 3.93, 1),
    (24, 'Cr', 5.64, 1),
    (25, 'Mn', 5.43, 1),
    (26, 'Fe', 7.50, 2),
    (27, 'Co', 4.99, 1),
    (28, 'Ni', 6.22, 1),
    (29, 'Cu', 4.19, 1),
    (30, 'Zn', 4.56, 1),
    (31, 'Ga', 3.04, 1),
    (32, 'Ge', 3.65, 1),
    (33, 'As', 2.30, 1),
    (34, 'Se', 3.34, 1),
    (35, 'Br', 2.54, 1),
    (36, 'Kr', 3.25, 1),
    (37, 'Rb', 2.36, 1),
    (38, 'Sr', 2.87, 1),
    (39, 'Y', 2.21, 1),
    (40, 'Zr', 2.58, 1),
    (41, 'Nb', 1.46, 1),
    (42, 'Mo', 1.88, 1),
    (43, 'Tc', 0, 0),
    (44, 'Ru', 1.75, 1),
    (45, 'Rh', 1.06, 1),
    (46, 'Pd', 1.65, 1),
    (47, 'Ag', 1.20, 1),
    (48, 'Cd', 1.71, 1),
    (49, 'In', 0.76, 1),
    (50, 'Sn', 2.04, 1),
    (51, 'Sb', 1.01, 1),
    (52, 'Te', 2.18, 1),
    (53, 'I', 1.55, 1),
    (54, 'Xe', 2.24, 1),
    (55, 'Cs', 1.08, 1),
    (56, 'Ba', 2.18, 1),
    (57, 'La', 1.10, 1),
    (58, 'Ce', 1.58, 1),
    (59, 'Pr', 0.72, 1),
    (60, 'Nd', 1.42, 1),
    (61, 'Pm', 0, 0),
    (62, 'Sm', 0.96, 1),
    (63, 'Eu', 0.52, 1),
    (64, 'Gd', 1.07, 1),
    (65, 'Tb', 0.30, 1),
    (66, 'Dy', 1.10, 1),
    (67, 'Ho', 0.48, 1),
    (68, 'Er', 0.92, 1),
    (69, 'Tm', 0.10, 1),
    (70, 'Yb', 0.92, 1),
    (71, 'Lu', 0.10, 1),
    (72, 'Hf', 0.85, 1),
    (73, 'Ta', -0.12, 1),
    (74, 'W',  0.65, 1),
    (75, 'Re', 0.26, 1),
    (76, 'Os', 1.40, 1),
    (77, 'Ir', 1.38, 1),
    (78, 'Pt', 1.62, 1),
    (79, 'Au', 0.80, 1),
    (80, 'Hg', 1.17, 1),
    (81, 'Tl', 0.77, 1),
    (82, 'Pb', 2.04, 1),
    (83, 'Bi', 0.65, 1),
    (84, 'Po', 0, 0),
    (85, 'At', 0, 0),
    (86, 'Rn', 0, 0),
    (87, 'Fr', 0, 0),
    (88, 'Ra', 0, 0),
    (89, 'Ca', 0, 0),
    (90, 'Th', 0.06, 1),
    (91, 'Pa', 0, 0),
    (92, 'U', -0.54, 1)
]


FORT5_HEADER = """%.2f %.2f
 T  F              ! LTE,  LTGRAY
 '%s'            ! name of file containing non-standard flags
*
* frequencies
*
 2000
*
* data for atoms   
*
 92                 ! NATOMS
* mode abn modpf
"""


FORT5_FOOTER = """*
* data for ions
*
*iat   iz   nlevs  ilast ilvlin  nonstd typion  filei
*
   1   -1     1      0      0      0    ' H- ' 'data/hmin.dat'
   1    0     9      0      0      0    ' H 1' 'data/h1.dat'
   1    1     1      1      0      0    ' H 2' ' '
   2    0    24      0      0      0    'He 1' 'data/he1.dat'
   2    1     1      1      0      0    'He 2' ' '
   6    0    40      0      0      0    ' C 1' 'data/c1.dat'
   6    1    22      0      0      0    ' C 2' 'data/c2.dat'
   6    2     1      1      0      0    ' C 3' ' '
   7    0    34      0      0      0    ' N 1' 'data/n1.dat'
   7    1    42      0      0      0    ' N 2' 'data/n2_32+10lev.dat'
   7    2     1      1      0      0    ' N 3' ' '
   8    0    33      0      0      0    ' O 1' 'data/o1_23+10lev.dat'
   8    1    48      0      0      0    ' O 2' 'data/o2_36+12lev.dat'
   8    2     1      1      0      0    ' O 3' ' '
  10    0    35      0      0      0    'Ne 1' 'data/ne1_23+12lev.dat'
  10    1    32      0      0      0    'Ne 2' 'data/ne2_23+9lev.dat'
  10    2     1      1      0      0    'Ne 3' ' '
  12    0    41      0      0      0    'Mg 1' 'data/mg1.t'
  12    1    25      0      0      0    'Mg 2' 'data/mg2.dat'
  12    2     1      1      0      0    'Mg 3' ' '
  13    0    10      0      0      0    'Al 1' 'data/al1.t'
  13    1    29      0      0      0    'Al 2' 'data/al2_20+9lev.dat'
  13    2     1      1      0      0    'Al 3' ' '
  14    0    45      0      0      0    'Si 1' 'data/si1.t'
  14    1    40      0      0      0    'Si 2' 'data/si2_36+4lev.dat'
  14    2     1      1      0      0    'Si 3' ' '
  16    0    41      0      0      0    ' S 1' 'data/s1.t'
  16    1    33      0      0      0    ' S 2' 'data/s2_23+10lev.dat'
  16    2     1      1      0      0    ' S 3' ' '
  26    0    30      0      0      0    'Fe 1' 'data/fe1.dat'
  26    1    36      0      0     -1    'Fe 2' 'data/fe2va.dat'
   0    0                                      'data/gf2601.gam'
                                               'data/gf2601.lin'
                                               'data/fe2p_14+11lev.rap'
  26    2    50      0      0     -1    'Fe 3' 'data/fe3va.dat'
   0    0                                      'data/gf2602.gam'
                                               'data/gf2602.lin'
                                               'data/fe3p_22+7lev.rap'
  26    3     1      1      0      0    'Fe 4' ' '
   0    0     0     -1      0      0    '    ' ' '
*
"""


FORT55 = """{imode:d}   {idsts:d}   {iprin:d}        ! imode idstd iprin
{inmod:d}    {intrpl:d}   {ichang:d}   {ichemc:d}     ! inmod intrpl ichang ichemc
{iophli:d}    {nunalp:d}   {nunbet:d}   {nungam:d}   {nunbal:d} ! iophli nunalp nunbet nungam nunbal
{ifreq:d}    {inlte:d}   {icontl:d}   {inlist:d}   {ifhe2:d} ! ifreq inlte icontl inlist ifhe2
{ihydpr:d}    {ihe1pr:d}   {ihe2pr}         ! ihydpr ihe1pr ihe2pr
{alam0:d}  {alast:d} {cutof0:d}  {cutofs:d}  {relop:g} {space:f} ! alam0 alast cutof0 cutofs relop space
1  20              ! nmlist, (iunitm(i),i=1,nmlist) for molecular linelists
"""


class SynspecGrid(Grid):
    """Synthesizes a new spectrum with Synspec at given grid positions."""

    def __init__(self, synspec: str, models: Grid, linelist: str, mollist: str, datadir: str,
                 range: Tuple[float, float], vturb: Union[str, float] = 2.0, elements: List[str] = None,
                 input: Union[str, Grid] = None, imode: int = 10, idstd: int = 0, iprin: int = 0, inmod: int = 0,
                 intrpl: int = 0, ichang: int = 0, ichemc: int = 1, iophli: int = 0, nunalp: int = 0, nunbet: int = 0,
                 nungam: int = 0, nunbal: int = 0, ifreq: int = 1, inlte: int = 0, icontl: int = 0, inlist: int = 0,
                 ifhe2: int = 0, ihydpr: int = 1, ihe1pr: int = 0, ihe2pr: int = 0, cutof0: int = 40, cutofs: int = 0,
                 relop: float = 1e-5, space: float = 0.03, normalize: bool = False, nstfile: str = 'nstf',
                 nd: int = None, ifmol: int = 1, tmolim: float = None, ippick: int = None, ibfac: int = None,
                 tempdir: str = None, solar_abund: Dict[str, float] = None, *args, **kwargs):
        """Constructs a new Grid.

        Args:
            synspec: Full path to synspec exectuble
            models: Grid with model atmospheres
            linelist: File with line list
            mollist: File with molecular list
            datadir: Name of data directory
            range: Tuple of start/end wavelenghts
            vturb: Either the microturbulence or a CSV file containing a table
            elements: List of elements to add as new axis
            input: Either the name of a fort.5 file or a Grid or None (in which case an automatic fort.5 will be used)
            parameters: Use this fort.55 file instead of the automatically generated one
            imode:
            idstd:
            iprin:
            inmod:
            intrpl:
            ichang:
            ichemc:
            iophli:
            nunalp:
            nunbet:
            nungam:
            nunbal:
            ifreq:
            inlte:
            icontl:
            inlist:
            ifhe2:
            ihydpr:
            ihe1pr:
            ihe2pr:
            cutof0:
            cutofs:
            relop:
            space:
            normalize: Normalize spectra
            nstfile: Name of file with non-standard flags
            nd:
            ifmol:
            tmolim:
            ippick:
            ibfac:
            tempdir: Temporary directory. Won't be delete if given.
            solar_abund: Dictionary with solar abundances to use.
        """
        from ..interpolator import Interpolator
        Grid.__init__(self, axes=None, *args, **kwargs)

        # store
        self._synspec = synspec
        self._linelist = linelist
        self._mollist = mollist
        self._datadir = datadir
        self._elements = [] if elements is None else elements
        self._range = range
        self._tempdir = tempdir
        self._parameters = dict(imode=imode, idsts=idstd, iprin=iprin, inmod=inmod, intrpl=intrpl, ichang=ichang,
                                ichemc=ichemc, iophli=iophli, nunalp=nunalp, nunbet=nunbet, nungam=nungam,
                                nunbal=nunbal, ifreq=ifreq, inlte=inlte, icontl=icontl, inlist=inlist, ifhe2=ifhe2,
                                ihydpr=ihydpr, ihe1pr=ihe1pr, ihe2pr=ihe2pr, alam0=range[0], alast=range[1],
                                cutof0=cutof0, cutofs=cutofs, relop=relop, space=space)
        self._normalize = normalize
        self._nstfile = nstfile
        self._nstf = dict(nd=nd, ifmol=ifmol, tmolim=tmolim, ippick=ippick, ibfac=ibfac)
        self._solar_abund = {} if solar_abund is None else solar_abund

        # load grid
        self._models: Grid = self.get_objects(models, [Grid, Interpolator], 'grids', self.log, single=True)

        # add and init axes
        self._axes = copy.deepcopy(self._models.axes())
        for el in self._elements:
            self._axes.append(GridAxis(name=el, initial=0.1, min=-10., max=10.))

        # vturb
        if isinstance(vturb, int) or isinstance(vturb, float):
            self._vturb = float(vturb)
        elif isinstance(vturb, str):
            filename = os.path.expandvars(vturb)
            self._vturb = pd.read_csv(filename, index_col=['Teff', 'logg', '[M/H]', '[alpha/M]'], dtype=float)

        # input/fort.5
        self._input = None
        if input is not None:
            # first check, whether this is a file
            if isinstance(input, str) and os.path.exists(input):
                # okay, take this
                self._input = input

            else:
                # try to create grid
                self._input: Grid = self.get_objects(input, [Grid, Interpolator], 'grids', self.log, single=True)

    def all(self) -> List[Tuple]:
        """Return all possible parameter combinations.

        Returns:
            All possible parameter combinations.
        """

        # get all params from model
        models = self._models.all()

        # vturb table?
        if isinstance(self._vturb, pd.DataFrame):
            return list(filter(lambda p: p in self._vturb.index, models))
        else:
            return models

    def __contains__(self, params: Tuple) -> bool:
        """Checks, whether the grid contains a given parameter set.

        Args:
            params: Parameter set to check.

        Returns:
            Whether or not the given parameter set exists in the grid.
        """
        if len(params) != len(self._axes):
            raise ValueError('Wrong number of parameters.')

        # check models
        if tuple(params[:4]) not in self._models:
            return False

        # check vturb
        if isinstance(self._vturb, pd.DataFrame) and tuple(params[:4]) not in self._vturb.index:
            return False

        # seems to exist
        return True

    def filename(self, params: Tuple) -> str:
        """Returns filename for given parameter set.

        Args:
            params: Parameter set to catch value for.

        Returns:
            Filename.
        """
        return None

    def __call__(self, params: Tuple) -> Any:
        """Fetches the value for the given parameter set.

        Args:
            params: Parameter set to catch value for.

        Returns:
            Grid value at given position.
        """

        # check
        if len(params) != len(self._axes):
            raise ValueError('Wrong number of parameters.')

        # get params
        # TODO: check, whether first four axes actually are the parameters we need
        teff, logg, feh, alpha = params[:4]
        changes = dict(zip(self._elements, params[4:]))

        # find element in models grid
        mod = self._models.filename(params[:4])

        # remember path
        cwd = os.getcwd()

        # go to temp directory
        tmp = self._tempdir if self._tempdir is not None else os.path.abspath(mkdtemp())
        os.chdir(tmp)

        # make sure to delete temp directory in the end
        try:
            # write fort.5 and file with non-standard flags
            self._write_fort5(teff, logg, feh, alpha)
            self._write_nstf(teff, logg, feh, alpha)

            # write config
            self._write_fort55()

            # write element changes
            self._write_fort56(changes, feh)

            # create symlinks
            os.symlink(os.path.expandvars(self._synspec), 'synspec')
            os.symlink(mod, 'fort.8')
            os.symlink(os.path.expandvars(self._linelist), 'fort.19')
            os.symlink(os.path.expandvars(self._mollist), 'fort.20')
            os.symlink(os.path.expandvars(self._datadir), 'data')

            # run synspec
            os.system('./synspec < fort.5 > fort.6')

            # read output
            d = pd.read_csv('fort.7', delim_whitespace=True, names=['wave', 'flux'])
            spec = Spectrum(wave=d['wave'], flux=d['flux']).resample_const(step=0.02)

            # normalize?
            if self._normalize:
                # read continuum and create
                c = pd.read_csv('fort.17', delim_whitespace=True, names=['wave', 'flux'])
                cont = Spectrum(wave=c['wave'], flux=c['flux']).resample(spec=spec)

                # divide
                spec.flux /= cont.flux

            # return it
            return spec

        finally:
            # return to old directory and clean up
            os.chdir(cwd)
            if self._tempdir is None:
                shutil.rmtree(tmp)

    def _write_fort5(self, teff, logg, feh, alpha):
        if self._input is not None:
            # single file or grid?
            if isinstance(self._input, str):
                # copy file
                shutil.copyfile(self._input, 'fort.5')
            else:
                # get from grid and copy
                filename = self._input.filename((teff, logg, feh, alpha))
                shutil.copyfile(filename, 'fort.5')

        else:
            # write automatically generated file
            with open('fort.5', 'w') as f:
                # write header
                f.write(FORT5_HEADER % (teff, logg, self._nstfile))

                # write abundances
                for no, el, abund, mode in ABUND_AGSS:
                    # calculate abundance
                    if mode == 0 or no == 0:
                        # H
                        a = 0
                    elif no == 1:
                        # He
                        a = 10.**(abund - 12)
                    elif no in [8, 10, 12, 14, 16, 18, 20, 22]:
                        # alpha elements
                        a = 10. ** (abund - 12 + feh + alpha)
                    else:
                        # other
                        a = 10. ** (abund - 12 + feh)

                    # write it
                    f.write('%d %.3g 0 !%s\n' % (mode, a, el))

                # write footer
                f.write(FORT5_FOOTER)

    def _write_nstf(self, teff, logg, feh, alpha_m):
        """Write file with non-standard flags."""

        # get vturb
        if isinstance(self._vturb, float):
            vturb = self._vturb
        elif isinstance(self._vturb, pd.DataFrame):
            vturb = float(self._vturb.loc[teff, logg, feh, alpha_m])
        else:
            return

        # write file
        with open(self._nstfile, 'w') as f:
            # write vturb
            f.write('VTB=%.2f\n' % vturb)
            # write parameters from constructor
            for key, val in self._nstf.items():
                # only write non-None values
                if val is not None:
                    f.write('%s=%s\n' % (key.upper(), str(val)))

    def _write_fort55(self):
        """Writes the fort.55 file."""

        # open file
        with open('fort.55', 'w') as f:
            # write config
            f.write(FORT55.format(**self._parameters))

    def _write_fort56(self, abunds: Dict[str, float], feh: float):
        """Write fort.56 file.

        To create the fort.56 that contains the updated abundance (abund_up) for the element wanted
        All we need is a file with :
            1
            iat abund_up

        Args:
            abunds: Dictionary as el->[X/H] abundance
        """

        with open('fort.56', 'w') as f:
            # write number of changes
            f.write('%d\n' % len(abunds))

            # write changes
            for user_el, user_abund in abunds.items():
                element_abundance = None
                element_number = None

                # given by user?
                if user_el in self._solar_abund:
                    # using value provided in c'tor
                    element_abundance = self._solar_abund[user_el]

                # find el in AGSS
                for no, el, abund, _ in ABUND_AGSS:
                    if el == user_el:
                        # need to set abundance?
                        if element_abundance is None:
                            element_abundance = abund

                        # element number
                        element_number = no

                        # found it, break here
                        break
                else:
                    raise ValueError('Element %s not found.' % user_el)

                # calculate abundance and write it
                a = 10. ** (element_abundance - 12 + user_abund + (feh if element_number > 1 else 0))
                f.write('%d %.3g\n' % (element_number, a))


__all__ = ['SynspecGrid']



import glob
from itertools import cycle

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib import gridspec
import numpy as np

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from spexxy.data import Spectrum

from spexxy.data import FitsSpectrum


def add_parser(subparsers):
    # add parser
    parser = subparsers.add_parser('plot', help='Plot one or more spectrum')
    parser.add_argument('spectra', type=str, help='Spectrum to plot', nargs='+')
    parser.add_argument('-o', '--output', help='Save plot to file', type=str)
    parser.add_argument('-r', '--results', help='Include results', action='store_true')
    parser.add_argument('-s', '--single', help='Plot all spectra into single plot', action='store_true')
    parser.add_argument('--range', type=float, nargs=2, help='Wavelength range to plot')
    parser.add_argument('--mode', type=str, choices=['wave', 'log'], help='Force wavelength mode for plot')
    parser.add_argument('--cmap', type=str, help='Color map to use for color cycle')
    parser.add_argument('--norm-to-mean', action='store_true', help='Normalize spectra to mean')

    # argparse wrapper for plot
    def run(args):
        # args to dict
        p = vars(args)

        # glob spectra
        spectra = []
        for s in p['spectra']:
            if '*' in s or '?' in s:
                spectra.extend(glob.glob(s))
            else:
                spectra.append(s)
        p['spectra'] = spectra

        # call method
        if args.single:
            plot_single(**p)
        else:
            plot(**p)
    parser.set_defaults(func=run)


def plot_single(spectra: list, mode: str = None, cmap: str = None, norm_to_mean: bool = False, **kwargs):
    from spexxy.data import Spectrum

    # create plot
    fig, ax = plt.subplots(figsize=(11.69, 8.27))

    # define color cycle
    if cmap is None:
        # take default color cycle
        gen = cycle(plt.rcParams['axes.prop_cycle'].by_key()['color'])
        # and repeat it to get N entries
        colors = [next(gen) for i in range(len(spectra))]
    else:
        # create new color cycle from cmap
        cmap = mpl.cm.get_cmap(cmap)
        colors = cmap(np.linspace(0.1, 0.9, len(spectra)))

    # loop spectra
    for i, filename in enumerate(sorted(spectra)):
        # load spectrum
        with FitsSpectrum(filename) as fs:
            # get spectrum
            spec = fs.spectrum

            # norm?
            if norm_to_mean:
                spec.norm_to_mean()

            # force mode?
            if mode:
                if mode == 'wave':
                    spec.mode(Spectrum.Mode.LAMBDA)
                else:
                    spec.mode(Spectrum.Mode.LOGLAMBDA)

            # plot it
            ax.plot(spec.wave, spec.flux, ls="-", lw=1., c=colors[i], marker="None", label=filename)

    # wrap up
    ax.set_xlabel(r"$\mathrm{Wavelength}\ \lambda\,[\mathrm{\AA}]$", fontsize=20)
    ax.set_ylabel('Flux', fontsize=20)
    ax.legend()
    ax.minorticks_on()
    ax.grid()
    fig.tight_layout()
    plt.show()


def plot(spectra: list, output: str = None, results: bool = False, range: list = None, mode: str = None,
         norm_to_mean: bool = False,**kwargs):
    from spexxy.data import Spectrum

    # if spectra not a list, make it a list
    if not isinstance(spectra, list):
        spectra = [spectra]

    # check output
    pdf = None
    if output:
        # want a PDF?
        if output.endswith('.pdf'):
            pdf = PdfPages(output)
        else:
            # if no pdf, we only allow for one spectrum to plot
            if len (spectra) > 1:
                raise ValueError('Plotting into a file other than a PDF works for a single spectrum only!')

    # loop spectra
    for filename in sorted(spectra):
        # load spectrum
        with FitsSpectrum(filename) as fs:
            # get spectrum
            spec = fs['NORMALIZED'] if 'NORMALIZED' in fs and results else fs.spectrum

            # norm?
            if norm_to_mean:
                spec.norm_to_mean()

            # and model
            model = fs.best_fit if results else None

            # get residuals
            residuals = fs.residuals if results else None

            # good pixels
            valid = fs.good_pixels

            # force mode?
            if mode:
                if mode == 'wave':
                    spec.mode(Spectrum.Mode.LAMBDA)
                else:
                    spec.mode(Spectrum.Mode.LOGLAMBDA)

        # plot
        plot_spectrum(spec, model, residuals, valid, wave_range=range, title=fs.filename)

        # show
        if output:
            pdf.savefig(orientation='landscape')
            plt.close()
        else:
            plt.show()

    # close file
    if output:
        pdf.close()


def plot_spectrum(spec, model: 'Spectrum' = None, residuals: np.ndarray = None, valid: np.ndarray = None,
                  wave_range: list = None, text: str = None, text_width: float = 0.3, title: str = None):
    # create plot
    fig = plt.figure(figsize=(11.69, 8.27))
    gs = gridspec.GridSpec(2, 1, figure=fig, height_ratios=[3, 1])
    gs.update(wspace=0., hspace=0., left=0.09, right=0.99, top=0.95, bottom=0.08)

    # adjust top to allow for a title
    if title is not None:
        gs.update(top=0.92)

    # adjust right to allow for some text
    if text is not None:
        gs.update(right=1.0 - text_width)

    # create figure and axes
    if residuals is None:
        ax_spectrum = plt.subplot(gs[:, 0])
        ax_residuals = None
    else:
        ax_spectrum = plt.subplot(gs[0, 0])
        ax_residuals = plt.subplot(gs[1, 0], sharex=ax_spectrum)

    # plot spectrum
    ax_spectrum.plot(spec.wave, spec.flux, ls="-", lw=1., c="k", marker="None")
    ax_spectrum.set_yticks(ax_spectrum.get_yticks()[1:-1])
    ax_spectrum.set_ylabel('Flux', fontsize=20)

    # best fit
    if model:
        ax_spectrum.plot(spec.wave, model, ls="-", lw=1., c="r", marker="None")

    # legend
    ax_spectrum.plot([0], [0], c='#F7977A')
    ax_spectrum.plot([0], [0], c='b')
    labels = ["Observation", "Model", "Mask", "Residuals"] if ax_residuals is not None else ["Observation", "Mask"]
    ax_spectrum.legend(labels, bbox_to_anchor=(0., 1.02, 1., .102), loc=3, ncol=4, mode="expand", borderaxespad=0.)

    # residuals
    if ax_residuals is not None:
        ax_residuals.plot(spec.wave, residuals, ls="-", lw=1., c="b", marker="None")
        ax_residuals.set_xlabel(r"$\mathrm{Wavelength}\ \lambda\,[\mathrm{\AA}]$", fontsize=20)
        ax_residuals.set_ylabel('Residuals', fontsize=20)
    else:
        ax_spectrum.set_xlabel(r"$\mathrm{Wavelength}\ \lambda\,[\mathrm{\AA}]$", fontsize=20)

    # xlim
    wave_min = wave_range[0] if wave_range else np.nanmin(spec.wave)
    wave_max = wave_range[1] if wave_range else np.nanmax(spec.wave)
    ax_spectrum.set_xlim(wave_min, wave_max)

    # ylim
    w = (spec.wave >= wave_min) & (spec.wave <= wave_max)
    if len(valid) == len(spec):
        w &= valid

    # get min/max of spectrum in range
    tmp = spec.flux[w]
    ax_spectrum.set_ylim((np.nanmin(tmp), np.nanmax(tmp)))

    # and for the residuals
    if ax_residuals is not None:
        tmp = residuals[w]
        ax_residuals.set_ylim((np.nanmin(tmp), np.nanmax(tmp)))

    # ticks and other stuff
    for ax in [ax_spectrum, ax_residuals]:
        if ax is not None:
            ax.minorticks_on()
            ax.grid()

            # good pixels
            start = None
            for x in range(len(spec.wave)):
                if valid[x] == 0 and start is None:
                    start = spec.wave[x]
                if valid[x] == 1 and start is not None:
                    # mark area
                    ax.axvspan(start - spec.wave_step / 2., spec.wave[x], color='#F7977A', zorder=10, alpha=0.5)
                    start = None
            if start is not None:
                # mark area
                ax.axvspan(start - spec.wave_step / 2., spec.wave[x], color='#F7977A', zorder=10, alpha=0.5)
                start = None

    # title
    if title is not None:
        fig.text(0.5, 0.97, title, ha='center', size=16)

    # add text
    fig.text(1.0 - text_width + 0.02, 0.9, text, family='monospace', va='top', ha='left')

    # finished
    return fig



import pandas as pd
import numpy as np

from .init import Init
from ..component import Component


class InitFromCsv(Init):
    """Initializes a component from a line in a CSV file.

    This class, when called, initializes the given parameters (or all if None) of a given component
    to the values in the given CSV file.
    """

    def __init__(self, filename: str = 'initials.csv', filename_col: str = 'Filename',
                 parameters: list = None, cmp_sep: str = ' ',
                 *args, **kwargs):
        """Initializes a new Init object.

        Args:
            filename: Name of CSV file.
            filename_col:  Name of column containing filename.
            parameters: List of parameter names to set from CSV.
            cmp_sep: String separating component and parameter name in CSV.
        """
        Init.__init__(self, *args, **kwargs)
        self._parameters = parameters
        self._cmp_sep = cmp_sep

        # load csv
        self.log.info('Reading CSV file with initial values from %s...', filename)
        self._csv = pd.read_csv(filename, index_col=filename_col)

    def __call__(self, cmp: Component, filename: str):
        """Initializes parameters of the given component with values from the CSV given in the configuration.

        Args:
            cmp: Component to initialize.
        filename: Filename of spectrum.
        """

        # got filename?
        try:
            self._csv.loc[filename]
        except KeyError:
            return

        # get lower case columns
        columns = {c.lower(): c for c in self._csv.columns}

        # parameters in component
        cmp_params = {c.lower(): c for c in cmp.param_names}

        # get list of parameters
        params = self._parameters if self._parameters is not None else cmp.param_names

        # and loop them
        for param in params:
            # lower case
            p = param.lower()

            # is it actually a parameter of the given component?
            if p in cmp_params:
                # find column for parameter
                col = None
                col_type = "initial"
                if cmp.prefix + self._cmp_sep + p in columns:
                    # <cmp>.<param> matches column name
                    col = columns[cmp.prefix + self._cmp_sep + p]
                elif p in columns:
                    # just <param> matches column name
                    col = columns[p]
                elif f"min({cmp.prefix}{self._cmp_sep}{p})" in columns:
                    # match min(<param>)
                    col = columns[f"min({cmp.prefix}{self._cmp_sep}{p})"]
                    col_type = 'min'
                elif f"min({p})" in columns:
                    # match min(<param>)
                    col = columns[f"min({p})"]
                    col_type = 'min'
                elif f"max({cmp.prefix}{self._cmp_sep}{p})" in columns:
                    # match max(<cmp>.<param>
                    col = columns[f"max({cmp.prefix}{self._cmp_sep}{p})"]
                    col_type = 'max'
                elif f"max({p})" in columns:
                    # match min(<param>)
                    col = columns[f"max({p})"]
                    col_type = 'max'

                # found column?
                if col is not None:
                    # get value
                    val = self._csv.at[filename, col]
                    if np.isnan(val):
                        val = None

                    # set it, if we got a valid value
                    if val is not None:
                        self.log.info(f'Setting {col_type} value for "{param}" of component "{cmp.prefix}" to {val}...')
                        c = 'value' if col_type == 'initial' else col_type
                        cmp.set(cmp_params[p], **{c: val})


__all__ = ['InitFromCsv']



from spexxy.data.lsf import LSF
from spexxy.data.spectrum import Spectrum


def add_parser(subparsers):
    # apply
    parser = subparsers.add_parser('apply')
    parser.set_defaults(func=run)
    parser.add_argument('lsf', metavar='lsf', type=str, help='LSF to convolve with')
    parser.add_argument('spectrum', metavar='spectrum', type=str, help='spectrum to convolve')
    parser.add_argument('output', metavar='output', type=str, help='output file')


def run(args):
    # load LSF and spec
    lsf = LSF.load(args.lsf)
    spec = Spectrum.load(args.spectrum)

    # scale
    lsf.resample(spec)

    # apply
    output = lsf(spec)

    # save it
    output.save(args.output)



from typing import List, Tuple

from . import Interpolator
from ..grid import Grid, GridAxis
from ..data import Spectrum


class LinearInterpolator(Interpolator):
    """A basic linear interpolator that operates on a given grid."""

    def __init__(self, grid: Grid, *args, **kwargs):
        """Initializes a new linear interpolator.

        Args:
            grid: Grid to interpolate on.
        """
        Interpolator.__init__(self, *args, **kwargs)

        # get grid
        self.log.info('Initializing linear interpolator...')
        self._grid: Grid = self.get_objects(grid, Grid, 'grids', single=True)
        self._axes: List[GridAxis] = self._grid.axes()

    @property
    def grid(self) -> Grid:
        """Returns grid used in this interpolator

        Returns:
            Grid used for this interpolator
        """
        return self._grid

    def axes(self) -> List[GridAxis]:
        """Returns information about the axes.

        Returns:
            List of GridAxis objects describing the grid's axes
        """
        return self._axes

    def __call__(self, params: Tuple) -> Spectrum:
        """Interpolates at the given parameter set

        Args:
            params: Parameter set to interpolate at

        Returns:
            Interpolated spectrum at given position
        """
        return self._interpolate(params)

    def _interpolate(self, params: Tuple, axis: int = None):
        """Do the actual interpolation at the given axis, then continue for axis-1.

        Args:
            params: Parameter set to interpolate at.
            axis: Axis to interpolate in - if None, we start at last available axis.

        Returns:
            Interpolated spectrum at given position.
        """

        # no axis given, start at latest
        if axis is None:
            axis = len(self._axes) - 1

        # check boundaries
        if params[axis] < self._axes[axis].min or params[axis] > self._axes[axis].max:
            raise KeyError('Requested parameters are outside the grid.')

        # let's get all possible values for the given axis
        axisValues = self._grid.axis_values(axis)

        # if params[axis] is on axis or axis is continuous, return it directly
        if axisValues is None or params[axis] in axisValues:
            if axis == 0:
                return self._grid(tuple(params))
            else:
                return self._interpolate(tuple(params), axis - 1)

        # on the last axis we enforce that the neighbour must exist
        p_lower = self._grid.neighbour(tuple(params), axis, 0, must_exist=axis == 0)
        p_higher = self._grid.neighbour(tuple(params), axis, 1, must_exist=axis == 0)
        if p_lower is None or p_higher is None:
            raise KeyError('No direct neighbours found in grid.')

        # get axis values
        x_lower = p_lower[axis]
        x_higher = p_higher[axis]

        # get data for p_lower and p_higher
        if axis == 0:
            lower_data = self._grid(p_lower)
            higher_data = self._grid(p_higher)
        else:
            lower_data = self._interpolate(p_lower)
            higher_data = self._interpolate(p_higher)

        # calculate coefficients
        A = (x_higher - params[axis]) / (x_higher - x_lower)
        B = 1. - A

        # return interpolation
        return lower_data * A + higher_data * B


__all__ = ['LinearInterpolator']



from typing import List, Tuple, Dict, Any

from . import Grid
from ..grid import GridAxis


class SubGrid(Grid):
    """A grid that only gives access to a part of the parent grid."""

    def __init__(self, parent: Grid, fixed: Dict[str, float], *args, **kwargs):
        """Initializes a new sub grid.

        Args:
            parent: Parent grid.
            fixed: Dictionary with fixed axis_name->value pairs.
        """

        # store
        self._parent = parent
        self._fixed = {}

        # get axes
        axes: List[GridAxis] = []
        for ax in parent.axes():
            if ax.name in fixed.keys():
                self._fixed[ax.name] = fixed[ax.name]
            else:
                axes.append(ax)

        # init Grid
        Grid.__init__(self, axes, *args, **kwargs)

        # log
        self.log.info('Initializing sub interpolator...')
        for ax, val in self._fixed.items():
            self.log.info('Fixed axis %s to value %f.', ax, val)

    def __call__(self, params: Tuple) -> Any:
        """Fetches the value for the given parameter set

        Args:
            params: Parameter set to catch value for.

        Returns:
            Grid value at given position.
        """

        # convert given params to list
        given_params = list(params)

        # loop all parameters
        full_params = []
        for ax in self._parent.axes():
            if ax.name in self._fixed.keys():
                # if axis is fixed, use fixed value
                full_params.append(self._fixed[ax.name])
            else:
                # otherwise pull next value from given_params
                full_params.append(given_params.pop(0))

        # call parent interpolator
        return self._parent(tuple(full_params))


__all__ = ['SubGrid']



import lmfit
import numpy as np
import scipy.linalg
from lmfit import Parameters
from lmfit.minimizer import MinimizerResult
from typing import List
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import pandas as pd
from spexxy.interpolator import Interpolator

from spexxy.data import FitsSpectrum, Spectrum
from spexxy.component import Component
from spexxy.mask import Mask
from spexxy.weight import Weight
from spexxy.data import SpectrumFitsHDU
from .base import FilesRoutine
from spexxy.tools.plot import plot_spectrum


class Legendre:
    """Calculate 'continuum' between a given spectrum and a model as a
    linear combination of Legendre polynomials."""

    def __init__(self, model, degree):
        """Initialize continuum fitter.

        :param model:   Model to use for deriving the continuum.
        :param degree:  Degree of Legendre polynomial to use.
        """

        # store variables
        self.model = model
        self.degree = degree
        self.coefficients = None
        self.mean = None
        self.values = np.ones((len(model)))

        # get wavelength array of model and normalize it to range -1..1
        # since we don't have mask here yet, this gets a little more complicated...
        wave = self.model.wave
        x = (wave - 0.5 * (wave[0] + wave[-1])) / 2.
        x /= np.max(np.abs(x[~np.isnan(x)]))

        # create Legendre polynomials
        # need to add 1 to polynomials to avoid zero points
        self._legendre = np.zeros((len(x), degree))
        for k in range(degree):
            self._legendre[~np.isnan(x), k] = scipy.special.legendre(k)(x[~np.isnan(x)]) + 1.

    def __call__(self, spec, valid=None, norm_to_mean=True):
        """Calculate the continuum.
        Effectively we're searching for the best coefficients c that
        fulfil S=c*A, where S is the flux of the spectrum and A is the
        matrix multiplication of all Legendre polynomials with the flux M of
        the model, i.e. A=L*M.t, where * denotes the matrix multiplication and
        M.t is the transposed of M.

        :param spec:    Spectrum to calculate continuum for.
        :param valid:   Array containing valid pixels in spectrum.
        :return:        numpy array containing continuum.
        """

        # check length
        if len(spec) != self._legendre.shape[0]:
            raise ValueError('Data length must be equal to initial data '
                             'length.')

        # valid
        if valid is None:
            valid = ~np.isnan(spec.flux)

        # add invalid model values to mask
        v = valid & ~np.isnan(self.model.flux)

        # fill matrix A as a matrix multiplication of L and the transposed
        # flux of the model
        A = np.empty((len(spec.flux[v]), self.degree))
        for k in range(self.degree):
            A[:, k] = spec.flux[v] * self._legendre[v, k]

        # solve, i.e. find best coefficients c for S=c*A
        self.coefficients = scipy.linalg.lstsq(A, self.model.flux[v])[0]

        # evaluate continuum
        self.values = np.dot(self._legendre, self.coefficients)

        # normalize to mean?
        if norm_to_mean:
            # divide evaluated polynomial by mean
            self.mean = self.values.mean()
            self.values /= self.mean

            # also divide first coefficient by mean
            self.coefficients[0] /= self.mean

        # return values
        return self.values


class ParamsFit(FilesRoutine):
    """ParamsFit is a fitting routine for spexxy that uses a Levenberg-Marquardt optimization to fit a set of
    model spectra (components) to a given spectrum.
    """

    def __init__(self, components: List[Component] = None, tellurics: Component = None, masks: List[Mask] = None,
                 weights: List[Weight] = None, fixparams: List[str] = None, poly_degree: int = 40,
                 maxfev: int = 500, ftol: float = 1.49012e-08, xtol: float = 1.49012e-08,
                 factor: float = 100.0, epsfcn: float = 1e-7, min_valid_pixels: float = 0.5,
                 plot_iterations: bool = False, check_limits: bool = True, *args, **kwargs):
        """Initialize a new ParamsFit object

        Args:
            components: List of components (or descriptions to create them) to fit to the spectrum
            tellurics: Tellurics component to add to the fit.
            masks:  List of mask objects to use for masking the spectrum.
            weights: List of weight objects to use for creating weights for the spectrum.
            fixparams: List of names of parameters to fix during the fit.
            poly_degree: Number of coefficients for the multiplicative polynomial.
            maxfev: The maximum number of calls to the function (see scipy documentation).
            ftol: Relative error desired in the sum of squares (see scipy documentation).
            xtol: Relative error desired in the approximate solution (see scipy documentation).
            factor: A parameter determining the initial step bound (factor * || diag * x||) (see scipy documentation).
            epsfcn: A variable used in determining a suitable step length for the forward- difference approximation of
                the Jacobian (see scipy documentation)
            min_valid_pixels: Fraction of minimum number of required pixels to continue with fit.
            plot_iterations: Plot all iterations into a PDF file.
            check_limits: Whether to check if results are close to limits.
        """
        FilesRoutine.__init__(self, *args, **kwargs)

        # remember variables
        self._max_fev = maxfev
        self._ftol = ftol
        self._xtol = xtol
        self._factor = factor
        self._epsfcn = epsfcn
        self._poly_degree = poly_degree
        self._fixparams = fixparams
        self._min_valid_pixels = min_valid_pixels
        self._check_limits = check_limits

        # spectrum
        self._spec = None
        self._valid = None
        self._weight = None

        # model
        self._model = None

        # polynomial
        self._mult_poly = None

        # iterations PDF?
        self._plot_iterations = plot_iterations
        self._iterations_pdf = None

        # find components
        self._cmps = self.get_objects(components, Component, 'components')

        # find tellurics
        self._tellurics = self.get_objects(tellurics, Component, 'components', single=True)

        # masks
        self._masks = self.get_objects(masks, Mask, 'masks')

        # weights
        self._weights = self.get_objects(weights, Weight, 'weights')

    def parameters(self) -> List[str]:
        """Get list of parameters fitted by this routine.

        Returns:
            List of parameter names (including prefix) fitted by this routine.
        """

        # init
        params = []

        # loop components
        for cmp in self._cmps:
            # add parameters
            params.extend(['%s %s' % (cmp.prefix, p) for p in cmp.param_names])

        # add tellurics
        if self._tellurics is not None:
            params.extend(['%s %s' % (self._tellurics.prefix, p) for p in self._tellurics.param_names])

        # finished
        return params

    def fit_parameters(self) -> List[str]:
        """Get list of parameters fitted by this routine. Exclude fixed parameters.

        Returns:
            List of parameter names (including prefix) fitted by this routine.
        """

        # init
        params = []

        # loop components
        for cmp in self._cmps:
            # loop all parameters of this component
            for param_name in cmp.param_names:
                # check if this parameter is fixed
                if self._fixparams and cmp.name in self._fixparams and param_name in self._fixparams[cmp.name]:
                    continue

                # add parameter to list
                params.append('{} {}'.format(cmp.prefix, param_name))

        return params

    def columns(self) -> List[str]:
        """Get list of columns returned by __call__.

        The returned list shoud include the list from parameters().

        Returns:
            List of columns returned by __call__.
        """

        # call base and add columns Success
        return FilesRoutine.columns(self) + ['Success', 'RedChi2']

    @property
    def components(self) -> List[Component]:
        """Returns all components used in this fit.

        Returns:
            List of all components.
        """
        return self._cmps + ([] if self._tellurics is None else [self._tellurics])

    def _fit(self, params: Parameters) -> MinimizerResult:
        """Perform the actual fit.

        Args:
            params: Parameters to fit.

        Returns:
            Result of fit
        """
        minimizer = lmfit.Minimizer(self._fit_func, params,
                                    iter_cb=self._callback, max_nfev=self._max_fev, nan_policy='raise',
                                    xtol=self._xtol, ftol=self._ftol, epsfcn=self._epsfcn, factor=self._factor)
        return minimizer.leastsq()

    def __call__(self, filename: str) -> List[float]:
        """Start the fitting procedure on the given file.

        Args:
            filename: Name of file to fit.

        Returns:
            List of final values of parameters, ordered in the same way as the return value of parameters()
        """

        # fix any parameters?
        for cmp_name, cmp in self.objects['components'].items():
            # loop all parameters of this component
            for param_name in cmp.param_names:
                # do we have parameters to fix and is this one of them?
                if self._fixparams and cmp_name in self._fixparams and param_name in self._fixparams[cmp_name]:
                    self.log.info('Fixing "%s" of component "%s" to its initial value of %f.',
                                  param_name, cmp_name, cmp[param_name])
                    cmp.set(param_name, vary=False)
                else:
                    # otherwise make it a free parameter
                    cmp.set(param_name, vary=True)

        # Load spectrum
        self._load_spectrum(filename)

        # create weight array
        self.log.info('Creating weights array...')
        self._weight = np.ones((len(self._spec)))
        if self._weights is not None:
            # loop all weights
            for w in self._weights:
                # multiply weights array with new weights
                self._weight *= w(self._spec, filename)

        # adjusting valid mask for weights
        self._valid &= ~np.isnan(self._weight)

        # less than 50% of pixels valid?
        if np.sum(self._valid) < self._min_valid_pixels * len(self._valid):
            self.log.warning('Less then %d percent of pixels valid, skipping...', self._min_valid_pixels * 100)
            return [None] * (len(self.columns()) - 2) + [False, 0]

        # initialize multiplicative polynomial with ones
        self._mult_poly = Legendre(self._spec, self._poly_degree)

        # get parameters
        params = Parameters()
        for cmp in self.components:
            params += cmp.make_params()

        # open PDF
        if self._plot_iterations:
            self._iterations_pdf = PdfPages(filename.replace('.fits', '.pdf'))

        # start minimization
        self.log.info('Starting fit...')
        result = self._fit(params)

        # close PDF file
        if self._plot_iterations:
            self._iterations_pdf.close()

        # get best fit
        best_fit = self._get_model(result.params)

        # estimate SNR
        snr = None if best_fit is None else self._spec.estimate_snr(best_fit)
        self.log.info('Estimated S/N of %.2f.', snr)

        # successful, if minimization was a success
        success = result.success

        # get message
        message = "" if result.lmdif_message is None else result.lmdif_message.replace("\n", " ")

        # if any of the parameters was fitted close to their edge, fit failed
        if self._check_limits:
            for pn in result.params:
                # ignore all sigma values
                if pn.lower().find("sig") != -1 or pn.lower().find("tellurics") != -1:
                    continue

                # get param
                p = result.params[pn]

                # get position of value within range for parameter
                pos = (p.value - p.min) / (p.max - p.min)

                # if pos < 0.01 or pos > 0.99, i.e. closer than 1% to the edge, fit failes
                if p.vary and (pos < 0.01 or pos > 0.99):
                    success = False
                    message = "Parameter %s out of range: %.2f" % (p.name, p.value)
                    break

        # fill statistics dict
        stats = {'success': success,
                 'errorbars': result.errorbars,
                 'nfev': result.nfev,
                 'chisqr': result.chisqr,
                 'redchi': result.redchi,
                 'nvarys': result.nvarys,
                 'ndata': result.ndata,
                 'nfree': result.nfree,
                 'msg': message,
                 'snr': snr}

        # write results back to file
        self._write_results_to_file(filename, result, best_fit, stats)

        # all components
        components = self._cmps
        if self._tellurics is not None:
            components.append(self._tellurics)

        # build list of results and return them
        results = []
        for cmp in self._cmps:
            # parse parameters
            cmp.parse_params(result.params)

            # loop params
            for n in cmp.param_names:
                p = '%s%s' % (cmp.prefix, n)
                results += [cmp.parameters[n]['value'], cmp.parameters[n]['stderr']]

        # success?
        results += [success, result.redchi]
        return results

    def _load_spectrum(self, filename: str) -> (Spectrum, np.ndarray):
        """Loads the given spectrum and its uncertainties and creates a mask.

        Args:
            filename: Name of file to load.

        Returns:
            Tuple of Spectrum and mask of valid pixels
        """

        # open file
        self.log.info("Loading file {0:s}.".format(filename))
        with FitsSpectrum(filename) as fs:
            # get spectrum
            self._spec = fs.spectrum

            # mask of good pixels
            self._valid = fs.good_pixels.astype(bool)

            # mask all NaNs
            self._valid &= ~np.isnan(self._spec.flux) & ~np.isnan(self._spec.wave)

            # add other masks
            for mask in self._masks:
                self._valid &= mask(fs.spectrum, filename=fs.filename)

    def _callback(self, params: Parameters, iter: int, resid: np.ndarray, *args, **kws):
        """Callback function that is called at each iteration of the optimization.

        Args:
            params: Current parameter values.
            iter: Current iteration.
            resid: Current residual array
        """
        # init
        messages = []

        # loop components
        for cmp in self._cmps + ([self._tellurics] if self._tellurics else []):
            values = []
            for name in cmp.param_names:
                # get param
                param = params[cmp.prefix + name]

                # de-normalize?
                p = cmp.parameters[name]
                val, _ = cmp.denorm_param(name, param.value) if cmp.normalize else (param.value, None)

                # show
                if param.vary:
                    values.append('%s=%.3f' % (name, val))

            # print it
            if len(values) > 0:
                messages.append('%s(%s)' % (cmp.prefix, ', '.join(values)))

        # calculate chi2
        sumsq = np.sum(((self._spec.flux[self._valid] - self._model.flux[self._valid])
                        * self._weight[self._valid])**2)
        chi2 = np.sum(((self._spec.flux[self._valid] - self._model.flux[self._valid])**2
                       / self._model.flux[self._valid]))
        messages += ['sqsum=%.5g' % sumsq, 'chi2=%.5g' % chi2]

        # plot spectrum
        if self._iterations_pdf is not None:
            self._plot_spectrum(params, iter)

        # log it
        self.log.info('(%3d) %s' % (iter, ', '.join(messages)))

    def _plot_spectrum(self, params: Parameters, iter: int):
        """Plot the current iteration.

        Args:
            params: Current parameter values.
            iter: Current iteration.
        """

        # build text
        text = 'Iteration %d\n===============\n' % iter

        # loop components and add current values
        for cmp in self._cmps + ([self._tellurics] if self._tellurics else []):
            text += '\n%s:\n' % cmp.prefix
            for name in cmp.param_names:
                param = params[cmp.prefix + name]
                text += '%10s = %.2f\n' % (name, param.value)

        # calculate chi2
        chi2 = np.sum(np.power(self._spec.flux[self._valid] - self._model.flux[self._valid], 2)
                      * self._weight[self._valid])
        text += '\nchi2 = %g\n' % chi2

        # do the plot
        fig = plot_spectrum(self._spec, model=self._model, residuals=self._spec.flux - self._model.flux,
                            valid=self._valid, title='Iteration %d' % iter, text=text)

        # store and close it
        self._iterations_pdf.savefig(fig)
        plt.close()

    def _fit_func(self, params: Parameters) -> np.ndarray:
        """Fit function for LM optimization

        Args:
            params: Parameters to evaluate.

        Returns:
            Weights scaled residuals between spec and model.
        """


        try:
            # evaluate params
            self._model = self._get_model(params)

            # calc residuals
            return (self._model.flux[self._valid] - self._spec.flux[self._valid]) * self._weight[self._valid]

        except (KeyError, pd.core.indexing.IndexingError, ValueError):
            # could not interpolate
            self.log.exception('Could not interpolate model.')
            self._model = Spectrum(spec=self._spec)
            self._model.flux[:] = 0
            return np.ones((len(self._spec.flux[self._valid]))) * 1e100

    def _get_model(self, params: Parameters) -> Spectrum:
        """Returns the model for the given parameters.

        Args:
            params: Parameters to evaluate.
            spec: Spectrum to fit.

        Returns:
            Model for given parameters.

        Raises:
            KeyError: Forwarded from grid/interpolator access.
            pd.core.indexing.IndexingError: Forwarded from grid/interpolator access.
        """

        # parse params
        for cmp in self.components:
            cmp.parse_params(params)

        # get tellurics
        tell = None
        if self._tellurics is not None:
            # evaluate component
            tell = self._tellurics()

            # resample to same wavelength grid as spec
            tell.mode(self._spec.wave_mode)
            tell = tell.resample(spec=self._spec)

        # get models for all components
        models = []
        for cmp in self._cmps:
            # evaluate model
            m = cmp()

            # all invalid?
            if len(m.flux) == len(m.flux[np.isnan(m.flux)]):
                raise ValueError('All values NaN.')

            # resample to same wavelength grid as spec
            m.mode(self._spec.wave_mode)
            m = m.resample(spec=self._spec)

            # multiply tellurics, if necessary
            if tell is not None:
                m.flux *= tell.flux

            # append to list
            models.append(m)

        # weight components
        if len(self._cmps) == 1:
            # get points that are valid in both model and spectrum
            v = self._valid & ~np.isnan(models[0].flux)

            # if we only have one component, it's easy :-)
            #a = models[0].flux[v] * self._mult_poly[v]
            a = models[0].flux[v] * self._mult_poly.values[v]
            b = self._spec.flux[v] * self._weight[v]
            self._cmps[0].weight = (a * b).sum() / (a * a).sum()
        else:
            self._fit_component_weights(models)

        # add all models together weighted
        model = models[0]
        model.flux *= self._cmps[0].weight
        for i in range(1, len(self._cmps)):
            model.flux += models[i].flux * self._cmps[i].weight

        # multiplicative poly
        cont = self._mult_poly(model, self._valid)
        cont_mean = self._mult_poly.mean

        # multiply weights with mean
        for c in self._cmps:
            c.weight /= cont_mean

        # multiply continuum
        model.flux *= cont
        model.flux *= cont_mean

        # return model
        return model

    def _calculate_mult_poly(self, model):
        # create Legendre fitter
        # obviously the Legendre module automatically maps the x values to the range -1..1
        leg = np.polynomial.Legendre.fit(self._spec.wave[self._valid],
                                         self._spec.flux[self._valid] / model.flux[self._valid],
                                         deg=self._poly_degree)

        # return new polynomial
        return leg(self._spec.wave)

    def _fit_component_weights(self, models: List[Spectrum]):
        """In case we got more than one model, we need to weight them.

        Args:
            models: List of models.
        """

        # get first model
        m0 = models[0]

        # get valid points
        valid = np.ones((len(m0)), dtype=bool)
        for m in models:
            valid &= ~np.isnan(m.flux)

        # create matrix with models
        mat = np.empty((len(m0[valid]), len(models)))
        for k in range(len(models)):
            mat[:, k] = models[k].flux[valid]

        # fit
        coeffs = scipy.linalg.lstsq(mat, self._spec.flux[valid])[0]

        # set weights
        for i, cmp in enumerate(self._cmps):
            cmp.weight = coeffs[i]

    def _write_results_to_file(self, filename: str, result: MinimizerResult, best_fit: Spectrum, stats: dict):
        """Writes results of fit back to file.

        Args:
            filename: Name of file to write results into.
            result: Result from optimization.
            best_fit: Best fit model.
            stats: Fit statistics.
        """

        # Write fits results back to file
        self.log.info("Writing results to file.")
        with FitsSpectrum(filename, 'rw') as fs:
            # stats
            res = fs.results('SPEXXY')
            for x in stats:
                res[x] = stats[x]

            # loop all components
            for cmp in self._cmps:
                # write results
                cmp.write_results_to_file(fs)

            # tellurics
            if self._tellurics is not None:
                # molecular abundances
                self._tellurics.write_results_to_file(fs)

            # weights
            weights = fs.results("WEIGHTS")
            for cmp in self._cmps:
                weights[cmp.prefix] = cmp.weight

            # write spectra best fit, good pixels mask, residuals and
            # multiplicative polynomial
            if best_fit is not None:
                fs.best_fit = best_fit
                fs.residuals = self._spec.flux - best_fit.flux
            fs.good_pixels = self._valid
            fs.mult_poly = self._mult_poly.values

            # loop all components again to add spectra
            for cmp in self._cmps:
                # get spectrum
                tmp = cmp()
                tmp.mode(self._spec.wave_mode)
                tmp = tmp.resample(spec=self._spec)
                cmpspec = SpectrumFitsHDU(spec=tmp, primary=False)

                # set it
                fs['CMP_' + cmp.name] = cmpspec

            # tellurics spectrum
            if self._tellurics is not None:
                tmp = self._tellurics()
                tmp.mode(self._spec.wave_mode)
                tmp = tmp.resample(spec=self._spec)
                tell = SpectrumFitsHDU(spec=tmp, primary=False)

                # set it
                fs['TELLURICS'] = tell

            # covariance
            if hasattr(result, 'covar'):
                fs.covar = result.covar


__all__ = ['ParamsFit']



import logging
import math
import os
from typing import List, Union
import astropy.io.fits as pyfits
import numpy as np
import pandas as pd
import scipy
import scipy.optimize
import scipy.stats
from astropy.convolution import convolve

from spexxy.data.losvd import LOSVD
from spexxy.data.spectrum import Spectrum


class LSF(object):
    """Base class for classes dealing with Line Spread Functions."""

    def wavelength_points(self, log: bool = False) -> List[float]:
        """List of wavelength points at which the LSF is defined.

        Args:
            log: Return the wavelength points as log(wavelength).

        Returns:
            List of wavelength points
        """
        raise NotImplementedError

    def __call__(self, spec: Spectrum) -> Spectrum:
        """Apply LSF to given spectrum and return result.

        Args:
            spec: Spectrum to apply LSF to.

        Returns:
            Final spectrum.
        """
        raise NotImplementedError

    def wave_mode(self, mode: Spectrum.Mode):
        """Change the wavelength mode of the LSF (LAMBDA or LOGLAMBDA)

        Args:
            mode: New wavelength mode for LSF.
        """
        raise NotImplementedError

    @staticmethod
    def load(filename: str) -> Union['EmpiricalLSF', 'AnalyticalLSF']:
        """Loads a LSF from file

        Args:
            filename: Name of LSF file.

        Returns:
            The LSF object.
        """

        # expand vars on filename
        filename = os.path.expandvars(filename)

        # if it's a FITS file, we assume it to be an empirical LSF
        if filename.upper().endswith('.FIT') or filename.upper().endswith('.FITS'):
            return EmpiricalLSF(filename)

        # if it's a TXT or CSV, it must be an analytical LSF
        if filename.upper().endswith('.CSV') or filename.upper().endswith('.TXT'):
            return AnalyticalLSF(filename)

        # otherwise we don't know
        raise ValueError("Unknown LSF format.")


class AnalyticalLSF(LSF):
    """Describes an analytical LSF, i.e. one that is defined as LOSVDs at different wavelengths."""

    def __init__(self, filename: str = None):
        """Initialize a new analutical LSF

        Args:
            filename: Name of file to load LSF from. If none, empty LSF is created.
        """
        LSF.__init__(self)

        # init
        self.data = None
        self._wave_mode = None

        # load from file?
        if filename:
            # read data file
            self.data = pd.read_csv(filename, index_col=False)

            # crude check for wavelength mode:
            # if first wavelength is <100, it's assumed to be on a log scale
            self._wave_mode = Spectrum.Mode.LAMBDA if self.data['wave'].iloc[0] > 100. else Spectrum.Mode.LOGLAMBDA

    def save(self, filename: str):
        """Save LSF to file.

        Args:
            filename: Name of file to save LSF into.
        """
        if self.data is not None:
            self.data.to_csv(filename, index=False)

    def wavelength_points(self, log: bool = False) -> list:
        """List of wavelength points at which the LSF is defined.

        Args:
            log: Return the wavelength points as log(wavelength).

        Returns:
            List of wavelength points.
        """

        # get wavelength array
        wave = sorted(self.data['wave'].values)

        # check wave mode and convert, if necessary
        if log and self.ctype2 == Spectrum.Mode.LAMBDA:
            return np.log(wave)
        elif not log and self.ctype2 == Spectrum.Mode.LOGLAMBDA:
            return np.exp(wave)
        else:
            return wave

    def wave_mode(self, mode: Spectrum.Mode):
        """Change the wavelength mode of the LSF (LAMBDA or LOGLAMBDA)

        Args:
            mode: New wavelength mode for LSF.
        """

        # no need to change anything?
        if mode == self._wave_mode:
            return

        # do the actual conversion
        if mode == Spectrum.Mode.LAMBDA:
            self.data['wave'] = np.exp(self.data['wave'])
        else:
            self.data['wave'] = np.log(self.data['wave'])

    def __call__(self, spec: Spectrum) -> Spectrum:
        """Apply LSF to given spectrum and return result.

        Args:
            spec: Spectrum to apply LSF to.

        Returns:
            Final spectrum.
        """

        # check
        if spec.wave_mode != Spectrum.Mode.LOGLAMBDA:
            raise ValueError("Spectrum must be sampled on log scale.")

        # create all losvds
        losvd = []
        for row in self.data[['wave', 'v', 'sig', 'h3', 'h4', 'h5', 'h6']].itertuples():
            l = LOSVD(row[2:])
            losvd.append(l.kernel(l.x(spec.wave_step)))

        # get maximum size of losvd
        losvd_npix = np.max([len(l) for l in losvd])

        # get half size of LSF
        npix = int(losvd_npix / 2.)

        # array for new flux
        new_flux = np.empty((len(spec.wave)))

        # calculate new flux
        for w in range(npix+1, len(new_flux)-npix):
            # find best lsf point
            lsf_idx = np.argmin(np.abs(spec.wave[w] - self.wavelength_points()))

            # half size of this specific lsf
            npix_lsf = int(len(losvd[lsf_idx]) / 2.)

            # calculate new flux
            new_flux[w] = np.sum(spec.flux[w-npix_lsf:w+npix_lsf+1] * losvd[lsf_idx]) * spec.wave_step

            # normalize
            new_flux[w] /= np.sum(losvd[lsf_idx]) * spec.wave_step

        # create new spec
        return spec.__class__(flux=new_flux[npix + 1:-npix], wave_start=spec.wave_start + (npix + 1) * spec.wave_step,
                              wave_step=spec.wave_step, wave_mode=spec.wave_mode)

    @staticmethod
    def from_empirical(lsf: 'EmpiricalLSF'):
        """Fits LOSVD profiles to a given empirical LSF and creates an analytical LSF from it.

        Args:
            lsf: LSF to create analytical LSF from.

        Returns:
            New analytical LSF.
        """

        # check
        if not isinstance(lsf, EmpiricalLSF):
            raise ValueError("LSF must be of type EmpiricalLSF.")

        # loop all wavelength point
        data = []
        for i, lam in enumerate(lsf.wavelength_points()):
            # get LSF for this wavelength and normalize it
            lsf_data = lsf.data[i, :]
            lsf_data /= np.sum(lsf_data)

            # wave step in km/s
            wave = 299792.458 * (np.exp(lsf.wave_lsf()) - 1.)

            # fit
            p0 = [0., 100., 0., 0., 0., 0.]
            p = scipy.optimize.leastsq(lambda p, x, y: LOSVD(p).kernel(x) - y,
                                       p0, args=(wave, lsf_data))

            # store results
            data.append([lam] + list(p[0]))

        # to pandas dataframe
        df = pd.DataFrame(data, columns=['wave', 'v', 'sig', 'h3', 'h4', 'h5', 'h6'])

        # create new LSF
        lsf_new = AnalyticalLSF()
        lsf_new.data = df
        return lsf_new

    def convolve(self, fwhm: float):
        """Convolve the LSF with a Gaussian of given FWHM.

        Args:
            fwhm: FWHM of Gaussian to convolve with.
        """

        # loop all wavelength points
        for i in range(len(self.data)):
            # get sigma
            sig = self.data['sig'].iloc[i]

            # (de)convolve, which is just a quadratic addition/subtraction for a Gaussian
            sig = np.sqrt(sig**2. + np.sign(fwhm)*fwhm**2.)

            # and set again
            self.data['sig'].iloc[i] = sig


class EmpiricalLSF(LSF):
    """Describes an empirical LSF, i.e. a measured one, where actual values are provided."""

    def __init__(self, filename: str = None):
        """Initialize a new empirical LSF.

        Args:
            filename: If given, load LSF from this file.
        """
        LSF.__init__(self)

        # init
        self.data = None
        self.crval1 = None
        self.cdelt1 = None
        self.crval2 = None
        self.cdelt2 = None
        self.ctype1 = None
        self.ctype2 = None

        # load from file?
        if filename is not None:
            # load LSF
            self.data, hdr = pyfits.getdata(filename, header=True)

            # get header info
            self.crval1 = hdr["CRVAL1"]
            self.cdelt1 = hdr["CDELT1"]
            self.crval2 = hdr["CRVAL2"]
            self.cdelt2 = hdr["CDELT2"]

            # wave mode
            self.ctype1 = Spectrum.Mode.LOGLAMBDA \
                if "CTYPE1" in hdr.keys() and hdr["CTYPE1"] == "WAVE-LOG" else Spectrum.Mode.LAMBDA
            self.ctype2 = Spectrum.Mode.LOGLAMBDA \
                if "CTYPE2" in hdr.keys() and hdr["CTYPE2"] == "WAVE-LOG" else Spectrum.Mode.LAMBDA

    def save(self, filename: str):
        """Save LSF to file.

        Args:
            filename: Name of file to write LSF into.
        """

        # create HDU and fill it
        hdu = pyfits.PrimaryHDU(self.data)
        hdu.header["CRVAL1"] = self.crval1
        hdu.header["CDELT1"] = self.cdelt1
        hdu.header["CD1_1"] = self.cdelt1
        hdu.header["CRPIX1"] = 1
        hdu.header["CRVAL2"] = self.crval2
        hdu.header["CDELT2"] = self.cdelt2
        hdu.header["CD2_2"] = self.cdelt2
        hdu.header["CRPIX2"] = 1
        hdu.header["CD1_2"] = 0.
        hdu.header["CD2_1"] = 0.
        hdu.header["CTYPE1"] = "WAVE" if self.ctype1 == Spectrum.Mode.LAMBDA else "WAVE-LOG"
        hdu.header["CTYPE2"] = "WAVE" if self.ctype2 == Spectrum.Mode.LAMBDA else "WAVE-LOG"

        # write it
        hdulist = pyfits.HDUList([hdu])
        if os.path.exists(filename):
            os.remove(filename)
        hdulist.writeto(filename)

    def wave_lsf(self):
        """Return wavelength array for the individual LSFs at different wavelengths."""
        return np.arange(self.crval1, self.crval1 + self.cdelt1 * self.data.shape[1], self.cdelt1)

    def wavelength_points(self, log: bool = False) -> List[float]:
        """List of wavelength points at which the LSF is defined.

        Args:
            log: Return the wavelength points as log(wavelength).

        Returns:
            List of wavelength points.
        """

        # build wave array
        wave = np.arange(self.crval2, self.crval2 + self.cdelt2 * self.data.shape[0], self.cdelt2)

        # check wave mode and convert, if necessary
        if log and self.ctype2 == Spectrum.Mode.LAMBDA:
            return np.log(wave)
        elif not log and self.ctype2 == Spectrum.Mode.LOGLAMBDA:
            return np.exp(wave)
        else:
            return wave

    def get_lsf_closest_to_wave(self, wave: float, log: bool = False) -> np.ndarray:
        """Return LSF defined closest to the given wavelength.

        Args:
            wave: Wavelength to return LSF for.
            log: Whether the given wavelength is on a log scale.

        Returns:
            LSF closest to given wavelength.
        """

        # get wavelength points
        points = np.array(self.wavelength_points(log=log))

        # get closest and return it
        p = np.argmin(abs(points - wave))
        return self.data[p, :]

    def extract_at_wave(self, wave: float) -> 'EmpiricalLSF':
        """Extract LSF at given wavelength

        Args:
            wave: Wavelength to extract at.

        Returns:
            New LSF
        """

        # create new lsf
        lsf = EmpiricalLSF()

        # copy values
        lsf.crval1 = self.crval1
        lsf.cdelt1 = self.cdelt1
        lsf.ctype1 = self.ctype1

        # ignore 2nd dimension
        lsf.crval2 = 0
        lsf.cdelt2 = 1
        lsf.ctype2 = Spectrum.Mode.LAMBDA

        # extract data and set it
        data = self.get_lsf_closest_to_wave(wave)
        lsf.data = data.reshape((1, len(data)))
        return lsf

    def wave_mode(self, mode: Spectrum.Mode):
        """Change the wavelength mode of the LSF (LAMBDA or LOGLAMBDA)

        Args:
            mode: New wavelength mode for LSF.
        """

        # no change required?
        if mode == self.ctype1:
            return

        # get wavelength array
        wave_lsf = self.wave_lsf()
        wave_spec = self.wavelength_points()

        # get largest extent of new array
        new_range = None
        for lam in wave_spec:
            # calculate new wavelength array for this lambda
            tmp = wave_lsf / lam if mode == Spectrum.Mode.LOGLAMBDA else wave_lsf * lam
            # larger?
            if new_range is None or tmp[-1]-tmp[0] > new_range[1]-new_range[0]:
                new_range = (tmp[0], tmp[-1])

        # create new wavelength array
        new_step = (new_range[1] - new_range[0]) / self.data.shape[1]
        new_wave_lsf = np.arange(new_range[0], new_range[1], new_step)

        # resample and rescale lsf
        for i, lam in enumerate(wave_spec):
            # get data
            data = self.data[i, :]

            # get wavelength array for this
            wave = wave_lsf / lam if mode == Spectrum.Mode.LOGLAMBDA else wave_lsf * lam

            # resample to new grid
            ip = scipy.interpolate.interp1d(x=wave, y=data, kind='linear', bounds_error=False, fill_value=0.)
            tmp = ip(new_wave_lsf)

            # normalize
            self.data[i, :] = tmp / np.sum(tmp) # np.trapz(tmp, new_wave_lsf)

        # set wave parameters
        self.crval1 = new_range[0]
        self.cdelt1 = new_step
        self.ctype1 = mode

    def resample(self, spec: Spectrum):
        """Resample LSF to match given spectrum.

        Args:
            spec: Spectrum to resample to.
        """

        # do wave modes match?
        if self.ctype1 != spec.wave_mode:
            raise ValueError("Wave modes of LSF and spectrum do not match.")

        # alreay good?
        if self.wave_lsf() == spec.wave:
            return

        # get number of pixels required for LSF and obtain new sampling
        npix = math.ceil(abs(self.crval1) / spec.wave_step)
        wave_lsf = np.arange(-spec.wave_step * npix, spec.wave_step * npix + spec.wave_step * 0.5, spec.wave_step)

        # blow up data to full spectrum's range and sampling
        if self.crval2 > 0:
            w = self.wavelength_points(log=(spec.wave_mode == Spectrum.Mode.LOGLAMBDA))
            ip = scipy.interpolate.interp2d(self.wave_lsf(), w, self.data)
            self.data = ip(wave_lsf, w)
        else:
            ip = scipy.interpolate.interp1d(self.wave_lsf(), self.data[0, :], fill_value='extrapolate')
            self.data = ip(wave_lsf).reshape(1, len(wave_lsf))

        # set wave parameters
        self.crval1 = wave_lsf[0]
        self.cdelt1 = spec.wave_step
        self.ctype1 = spec.wave_mode

    def __call__(self, spec: Spectrum) -> Spectrum:
        """Apply LSF to given spectrum and return result.

        Args:
            spec: Spectrum to apply LSF to.

        Returns:
            Final spectrum.
        """

        # check
        if self.cdelt1 != spec.wave_step:
            raise ValueError("Samplings of LSF and spectrum do not match.")

        # get lsf wave points
        wave_spec = self.wavelength_points(spec.wave_mode == Spectrum.Mode.LOGLAMBDA)

        # get half size of LSF
        npix = int(self.data.shape[1] / 2.)

        # LSF at multiple wavelengths?
        if self.crval2 > 0:
            # array for new flux
            new_flux = np.empty((len(spec.wave)))

            # calculate new flux
            for w in range(npix+1, len(new_flux)-npix):
                # find best lsf point
                lsf_idx = np.argmin(np.abs(spec.wave[w] - wave_spec))
                # calculate new flux
                new_flux[w] = np.sum(spec.flux[w-npix:w+npix+1] * self.data[lsf_idx, :]) # * spec.wave_step
                new_flux[w] /= np.sum(self.data[lsf_idx, :])
        else:
            # just convolve
            new_flux = convolve(spec.flux, self.data[0, :], boundary='extend')

        # create new spec
        return spec.__class__(flux=new_flux[npix + 1:-npix], wave_start=spec.wave_start + (npix + 1) * spec.wave_step,
                              wave_step=spec.wave_step, wave_mode=spec.wave_mode)


__all__ = ['LSF', 'AnalyticalLSF', 'EmpiricalLSF']



import logging
import argparse
import os

from spexxy.grid import Grid
from spexxy.data import LSF, Spectrum, SpectrumFits

log = logging.getLogger(__name__)


def add_parser(subparsers):
    # create parser
    parser = subparsers.add_parser('convolve', help='Convolves all spectra in a grid',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('input', type=str, help='Input grid')
    parser.add_argument('output', type=str, help='Output directory, in which a file grid is written')
    parser.add_argument('-w', '--wave', type=float, nargs=3, help='Start, end and step value for the wavelengths grid',
                        required=True)
    parser.add_argument('-o', '--overwrite', action="store_true", help='Overwrite existing files.')
    parser.add_argument('-l', '--log', action="store_true",
                        help='Create spectra on log scale, now -w and -f are given in log units.')
    parser.add_argument('-c', '--logconv', action="store_true", help='Do convolution on logarithmic scale.')
    parser.add_argument('-a', '--air', action="store_true", help='Convert spectra to air wavelengths.')
    parser.add_argument('-n', '--normalize', action="store_true", help='Normalize spectra to a mean flux of 1.')
    parser.add_argument('-p', '--precrop', type=float, nargs=2, help='Crop spectrum to this range before doing anything.')
    parser.add_argument('--spec-loader', type=str, help='Class for loading spectra.')

    # group for fwhm and lsf
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-f', '--fwhm', metavar='fwhm', type=float, help='FWHM of gaussian to convolve spectra with')
    group.add_argument('-v', '--lsf', metavar='file', type=str, help='Name of file containing LSF')

    # argparse wrapper for create_grid
    def run(args):
        convolve_grid(args.input, args.output, args.wave[0], args.wave[1], args.wave[2],
                      overwrite=args.overwrite, vac2air=args.air, log_scale=args.log, log_convolve=args.logconv,
                      normalize=args.normalize, fwhm=args.fwhm, lsf_file=args.lsf, precrop=args.precrop,
                      spec_loader=args.spec_loader)
    parser.set_defaults(func=run)


def convolve_grid(ingrid: str, outdir: str, wave_start: float, wave_end: float, sampling: float,
                  overwrite: bool = True, vac2air: bool = True, log_scale: bool = False,
                  log_convolve: bool = False, normalize: bool = False, fwhm: float = None,
                  lsf_file: str = None, precrop = None, spec_loader: str = None):

    # load grid
    grid = Grid.load(ingrid, spec_loader=spec_loader)

    # outdir and grid file
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    with open(os.path.join(outdir, 'grid.csv'), 'w') as f:
        f.write('Filename,' + ','.join(grid.axis_names()) + '\n')

    # get all params
    all_params = grid.all()

    # loop params
    for i, params in enumerate(all_params):
        log.info('(%d/%d) %s' % (i, len(all_params), params))

        # get spectrum
        log.info("  Loading spectrum...")
        spec = grid(params)

        # does spectrum have a filename or do we need to create one?
        if hasattr(spec, 'filename'):
            filename = spec.filename
            basepath = os.path.relpath(filename, os.path.dirname(ingrid))
        else:
            filename = 'spec_' + '_'.join(['%.2f' % p for p in params]) + '.fits'
            basepath = filename

        # get name and directory of output file
        outfile = os.path.join(outdir, basepath)
        outpath = os.path.dirname(outfile)

        # does dir exist?
        if not os.path.exists(outpath):
            os.makedirs(outpath)

        # exists?
        if os.path.exists(outfile) and not overwrite:
            log.info('  Exists, skipping...')
            continue

        # cropping it?
        if precrop is not None:
            spec = spec.extract(precrop[0], precrop[1])

        # log? does nothing, if modes already match
        spec.mode(Spectrum.Mode.LOGLAMBDA if log_convolve else Spectrum.Mode.LAMBDA)

        # process spec
        if fwhm:
            # constant sampling
            if spec.wave_step == 0.:
                spec = spec.resample_const()
            # convolve
            log.info("  Convolving with FWHM=%f...", fwhm)
            spec.smooth(fwhm)
        elif lsf_file is not None:
            # constant sampling
            if spec.wave_step == 0.:
                spec = spec.resample_const()

            # load lsf and resample
            log.info("  Loading and resampling LSF...")
            lsf = LSF.load(lsf_file)
            lsf.wave_mode(spec.wave_mode)
            lsf.resample(spec)

            # convolve
            log.info("  Convolving with LSF...")
            spec = lsf(spec)

        # vac2air
        if vac2air:
            log.info("  Converting from vacuum to air wavelengths...")
            spec.vac_to_air()

        # log? both commands do nothing, if modes already match
        spec.mode(Spectrum.Mode.LOGLAMBDA if log_scale else Spectrum.Mode.LAMBDA)

        # resample
        wave_count = int((wave_end - wave_start) / sampling + 1.)
        log.info("  Resampling to final grid, wave_start=%f, wave_step=%f, wave_count=%d...",
                 wave_start, sampling, wave_count)
        spec = spec.resample(wave_start=wave_start, wave_step=sampling, wave_count=wave_count)

        # normalize?
        if normalize:
            spec.norm_to_mean()

        # save
        log.info("  Saving to {0:s}...".format(outfile))
        SpectrumFits(spec=spec).save(outfile)

        # add to CSV
        with open(os.path.join(outdir, 'grid.csv'), 'a') as csv:
            csv.write(filename + ',' + ','.join([str(p) for p in params]) + '\n')



from typing import List, Tuple, Dict
import numpy as np

from .base import MainRoutine
from ..data import SpectrumFits, Spectrum, LSF
from ..grid import Grid


class FromGrid(MainRoutine):
    """Grid is a main routine for fetching a spectrum from a grid."""

    def __init__(self, grid: Grid, params: Dict[str, Tuple], prefix: str = 'spectrum_', lsf: str = None,
                 *args, **kwargs):
        """Initialize a new ParamsFit object

        Args:
            grid: Grid to extract from.
            param: Dict with filename->parameters pairs with spectra to extract.
            prefix: Prefix for spectra to write.
            lsf: Name of file containing LSF.
        """
        MainRoutine.__init__(self, *args, **kwargs)
        self._grid = grid
        self._params = params
        self._prefix = prefix
        self._lsf = lsf

    def __call__(self):
        """Start the routine."""

        # get grid
        grid = self.get_objects(self._grid, Grid, 'grids', self.log, single=True)

        # loop params
        for i, (filename, params) in enumerate(self._params.items(), 1):
            # get spectrum
            self.log.info('Fetching spectrum at %s...', params)
            spec: Spectrum = grid(tuple(params))

            # lsf?
            if self._lsf is not None:
                # resample const
                spec = spec.resample_const()

                # load lsf
                lsf = LSF.load(self._lsf).extract_at_wave(np.mean(spec.wave))
                lsf.resample(spec)

                # apply it
                spec = lsf(spec)

            # write it
            self.log.info('Saving as %s...', filename)
            SpectrumFits(spec=spec).save(filename)


__all__ = ['FromGrid']



from typing import List, Dict, Any, Tuple

from .grid import Grid, GridAxis


class ValuesGrid(Grid):
    """Basic Grid that gets its values from the constructor. Mainly used for testing."""

    def __init__(self, axes: List[GridAxis], values: Dict[Tuple, Any], *args, **kwargs):
        """Constructs a new Grid.

        Args:
            axes: A list of axes to build the grid from.
            values: A dictionary where the keys are the parameters and the values the values at that
                position in the grid.
        """
        Grid.__init__(self, axes, *args, **kwargs)

        # remember values
        self._values = values

    def all(self) -> List[Tuple]:
        """Return all possible parameter combinations.

        Returns:
            All possible parameter combinations.
        """
        return list(self._values.keys())

    def __contains__(self, params: Tuple) -> bool:
        """Checks, whether the grid contains a given parameter set.

        Args:
            params: Parameter set to check.

        Returns:
            Whether or not the given parameter set exists in the grid.
        """
        return tuple(params) in self._values

    def __call__(self, params: Tuple) -> Any:
        """Fetches the value for the given parameter set

        Args:
            params: Parameter set to catch value for

        Returns:
            Grid value at given position
        """
        return self._values[params]


__all__ = ['ValuesGrid']



import logging
import argparse
import os
import numpy as np
from scipy.interpolate import UnivariateSpline

from spexxy.grid import Grid
from spexxy.data import SpectrumFits
from spexxy.interpolator.spline import calc_2nd_derivs_spline

log = logging.getLogger(__name__)


def add_parser(subparsers):
    # create parser
    parser = subparsers.add_parser('2ndderivs', help='Calculate 2nd derivativates for a given grid along its 1st axis',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('input', type=str, help='Input grid')
    parser.add_argument('output', type=str, help='Output directory, in which a file grid is written')
    parser.add_argument('--norm-to-mean', action='store_true', help='Norm input spectra to mean.')

    # argparse wrapper for create_grid
    def run(args):
        calc_2nd_derivs(args.input, args.output, args.norm_to_mean)
    parser.set_defaults(func=run)


def calc_2nd_derivs(ingrid: str, outdir: str, norm_to_mean: bool = False):
    # load grid
    grid = Grid.load(ingrid)

    # outdir and grid file
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    with open(os.path.join(outdir, 'grid.csv'), 'w') as f:
        f.write('Filename,' + ','.join(grid.axis_names()) + '\n')

    # get first axis
    axis = grid.axes()[0]
    log.info('First axis is %s with %d values.', axis.name, len(axis.values))

    # get all parameter tuples
    all_params = grid.all()

    # load first spectrum to get wavelength array
    ref_spec = grid(all_params[0])
    log.info('Spectra in grid contain %d wavelength points.', len(ref_spec.wave))

    # get all parameter combinations without first parameter
    filtered_params = list(set([tuple(p[1:]) for p in all_params]))
    log.info('Found %d parameter combinations excluding 1st parameter.', len(filtered_params))

    # loop filtered params
    for i, params in enumerate(filtered_params, 1):
        log.info('(%d/%d) Calculating 2nd derivatives for %s...', i, len(filtered_params),
                 ' '.join(['%s=%.2f' % (k, v) for k, v in zip(grid.axis_names()[1:], params)]))

        # load all spectra with these parameters
        data = []
        avail_values = []
        for value in axis.values:
            try:
                # load spectrum
                spec = grid(tuple([value] + list(params)))

                # norm?
                if norm_to_mean:
                    spec.norm_to_mean()

                # append flux to data
                data.append(spec.flux)
                avail_values.append(value)

            except KeyError:
                # could not load spectrum
                continue

        # log
        log.info('Found %d different values for %s.', len(avail_values), axis.name)

        # calculate 2nd derivatives from spline
        derivs = calc_2nd_derivs_spline(avail_values, data)

        # loop spectra
        log.info('Writing spectra...')
        for i, value in enumerate(avail_values):
            # create spectrum
            spec = SpectrumFits(spec=ref_spec, flux=derivs[i, :])

            # create filename
            filename = 'spec_' + '_'.join(['%.2f' % p for p in [value] + list(params)]) + '.fits'

            # save it
            spec.save(os.path.join(outdir, filename))

            # add to CSV
            with open(os.path.join(outdir, 'grid.csv'), 'a') as csv:
                csv.write(filename + ',' + str(value) + ',' + ','.join([str(p) for p in params]) + '\n')



from typing import List
import numpy as np
import pandas as pd
import scipy.special
import scipy.linalg
import scipy.optimize
from spexxy.data import Spectrum, FitsSpectrum
import matplotlib.pyplot as plt

from .base import FilesRoutine
from ..grid import FitsGrid


class LinearCombinationFit(FilesRoutine):
    """A routine that tries to fit a given spectrum by finding a linear combination of spectra in a grid."""

    def __init__(self, grid: FitsGrid, *args, **kwargs):
        """Initialize a new LinearCombinationFit object

        Args:
            grid: Grid to use for fitting.
        """
        FilesRoutine.__init__(self, *args, **kwargs)

        # create grid
        self._grid: FitsGrid = self.get_objects(grid, FitsGrid, 'grids', single=True)
        self._poly_degree = 5

    def parameters(self) -> List[str]:
        """Get list of parameters fitted by this routine.

        Returns:
            List of parameter names (including prefix) fitted by this routine.
        """

        # just a list of grid axes
        return ['STAR %s' % n for n in self._grid.axis_names()]

    def __call__(self, filename: str) -> List[float]:
        """Start the fitting procedure on the given file.

        Args:
            filename: Name of file to fit.

        Returns:
            List of final values of parameters, ordered in the same way as the return value of parameters()
        """

        # get data from grid
        grid = self._grid.data

        # create Legendre polynomials
        legendre = np.empty((self._poly_degree, grid.shape[1]))
        x = np.linspace(-1, 1, grid.shape[1])
        for n in range(self._poly_degree):
            leg = scipy.special.legendre(n)
            legendre[n, :] = leg(x)

        # concatenate arrays
        data = np.concatenate((grid, legendre), axis=0)

        # load spectrum
        spec = Spectrum.load(filename)
        mask = ~np.isnan(spec.flux)

        # do fit
        x, _ = scipy.optimize.nnls(data[:, mask].T, spec.flux[mask])

        # get bestfit
        best_fit = Spectrum(spec=spec)
        best_fit.flux = data.T @ x

        # get table with parameters and weights
        params = dict(zip(self._grid.axis_names(), list(map(list, zip(*self._grid.all())))))
        params['weight'] = x[:-self._poly_degree] / np.sum(x)
        weights_table = pd.DataFrame(params)

        # write results
        self._write_results_to_file(filename, spec, weights_table, best_fit)

        # do weighting in table
        results = []
        for col in self._grid.axis_names():
            # calculate weighted mean and append it and error
            mean = np.sum(weights_table[col] * weights_table['weight'])
            results.extend([mean, 0.])

        # finished
        return results

    def _write_results_to_file(self, filename: str, spec: Spectrum, weights: pd.DataFrame, best_fit: Spectrum):
        """Writes results of fit back to file.

        Args:
            filename: Name of file to write results into.
            weights: Weights array.
            best_fit: Best fit model.
        """

        # Write fits results back to file
        self.log.info("Writing results to file.")
        with FitsSpectrum(filename, 'rw') as fs:
            # write spectra best fit and residuals
            if best_fit is not None:
                fs.best_fit = best_fit
                fs.residuals = spec.flux - best_fit.flux

        # write weights (as csv for now)
        weights.to_csv(filename.replace('.fits', '.csv'), index=False)


__all__ = ['LinearCombinationFit']



import logging
import argparse
import sys
from typing import List, Tuple, Dict
import yaml

from spexxy.component import SpectrumComponent
from spexxy.data import Spectrum, SpectrumFits
from spexxy.grid import Grid
from spexxy.interpolator import Interpolator
from spexxy.object import create_object

log = logging.getLogger(__name__)


def key_value_pair_or_value(opt: str):
    """Argument type for argparse, which accepts either a single float or a key=value pair.

    Args:
        opt: Command line parameter

    Returns:
        Single float or tuple containing key and value
    """

    # does it contain a =?
    pos = opt.find('=')
    if pos < 0:
        # single value?
        try:
            return float(opt)
        except ValueError:
            raise argparse.ArgumentTypeError('Value %s is not a float.' % opt)

    # split
    key = opt[:pos]
    value = opt[pos + 1:]

    # convert value to float
    try:
        value = float(value)
    except ValueError:
        raise argparse.ArgumentTypeError('Value %s for parameter %s is not a float.' % (value, key))

    # and return
    return key, value


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('get', help='Get a spectrum from a grid or interpolator',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('params', type=key_value_pair_or_value, nargs='+',
                        help='Parameters for spectrum as list of values or key=value pairs')
    parser.add_argument('-o', '--output', type=str, help='Output fits file', default='spectrum.fits')
    parser.add_argument('-c', '--config', type=str, help='Config for grid', default='spexxy.yaml')

    # argparse wrapper for create_grid
    parser.set_defaults(func=lambda args: get_spectrum_from_grid(**vars(args)))


def check_params(names: List[str], params: Dict[str, float]):
    """Checks whether all names show up in params and vice versa.

    Args:
        names: List of parameter names from grid/interpolator/component
        params: Dictionary with parameters from command line

    Raises:
        argparse.ArgumentTypeError if lists don't match
    """
    if sorted(names) != sorted(params.keys()):
        raise argparse.ArgumentTypeError('Given list of parameters does not match parameters '
                                         'in grid/interpolator/component.')


def get_spectrum_from_grid(params: List[Tuple[str, float]], output: str, config: str, **kwargs):
    """Fetches a spectrum with given parameters from the grid/interpolator/component given in
    config and stores it in output.

    Args:
        params: Parameters for spectrum
        output: Output filename
        config: Configuration for grid/interpolator/component
    """

    # list of values or list of tuples?
    if all([isinstance(p, float) for p in params]):
        # all floats
        pass
    elif all(isinstance(p, tuple) for p in params):
        # params to dict
        params = dict(params)
    else:
        raise argparse.ArgumentTypeError('Parameter list must either be a list of floats or of param=value pairs.')

    # create grid
    log.info('Initializing grid/interpolator/component...')
    with open(config, 'r') as f:
        cfg = yaml.safe_load(f)
    obj = create_object(config=cfg, log=log)

    # what type is it?
    if isinstance(obj, Grid) or isinstance(obj, Interpolator):
        # log
        log.info('Created an instance of %s, which is %s.',
                 type(obj).__name__, 'a Grid' if isinstance(obj, Grid) else 'an Interpolator')

        # need to make dict?
        if isinstance(params, list):
            # check length of list
            if len(params) != len(obj.axes()):
                raise argparse.ArgumentTypeError('Invalid number of parameters')

            # to dict
            params = dict(zip([ax.name for ax in obj.axes()], params))

        # check params
        check_params([ax.name for ax in obj.axes()], params)

        # get parameters
        p = [params[ax.name] for ax in obj.axes()]

        # get spectrum
        log.info('Fetching spectrum at %s...', ', '.join(['%s=%.2f' % x for x in params.items()]))
        spec: Spectrum = obj(tuple(p))

    elif isinstance(obj, SpectrumComponent):
        # log
        log.info('Created an instance of %s, which is a SpectrumComponent.', type(obj).__name__)

        # need to make dict?
        if isinstance(params, list):
            # check length of list
            if len(params) != len(obj.param_names):
                raise argparse.ArgumentTypeError('Invalid number of parameters')

            # to dict
            params = dict(zip(obj.param_names, params))

        # check params
        check_params(obj.param_names, params)

        # set params
        for param, value in params.items():
            obj.set(param, value=value)

        # get spectrum
        log.info('Fetching spectrum at %s...', ', '.join(['%s=%.2f' % x for x in params.items()]))
        spec = obj()

    else:
        log.error('Created object is not of any supported type.')
        sys.exit(1)

    # write it
    log.info('Saving as %s...', output)
    SpectrumFits(spec=spec).save(output)
    log.info('Done.')



import logging
import argparse
import os

from spexxy.grid import Grid
from spexxy.data import LSF, Spectrum, SpectrumFits

log = logging.getLogger(__name__)


def add_parser(subparsers):
    # create parser
    parser = subparsers.add_parser('resample', help='Resample all spectra in a grid',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.set_defaults(func=run)
    parser.add_argument('input', type=str, help='Input grid')
    parser.add_argument('output', type=str, help='Output directory, in which a file grid is written')
    parser.add_argument('wave_start', type=float, help='Start of the wavelengths grid')
    parser.add_argument('wave_step', type=float, help='Step size for the wavelengths grid')
    parser.add_argument('wave_count', type=int, help='Length of the wavelengths grid')
    parser.add_argument('-m', '--mode', type=str, choices=['lambda', 'log'], default='lambda',
                        help='Target wavelength mode')


def run(args):
    resample_grid(args.input, args.output, args.wave_start, args.wave_step, args.wave_count,
                  Spectrum.Mode[args.mode.upper()])


def resample_grid(ingrid: str, outdir: str, wave_start: float, wave_step: float, wave_count: float,
                  mode: Spectrum.Mode):

    # load grid
    grid = Grid.load(ingrid)

    # outdir and grid file
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    with open(os.path.join(outdir, 'grid.csv'), 'w') as csv:
        csv.write('Filename,' + ','.join(grid.axis_names()) + '\n')

    # get all params
    all_params = grid.all()

    # loop params
    for i, params in enumerate(all_params):
        log.info('(%d/%d) %s' % (i, len(all_params), params))

        # get spectrum
        spec = grid(params)

        # does spectrum have a filename or do we need to create one?
        if hasattr(spec, 'filename'):
            filename = spec.filename
            basepath = os.path.relpath(filename, os.path.dirname(ingrid))
        else:
            filename = 'spec_' + '_'.join(['%.2f' % p for p in params]) + '.fits'
            basepath = filename

        # get name and directory of output file
        outfile = os.path.join(outdir, basepath)
        outpath = os.path.dirname(outfile)

        # does dir exist?
        if not os.path.exists(outpath):
            os.makedirs(outpath)

        # change mode
        spec.mode(mode)

        # resample it
        resampled = spec.resample(wave_start=wave_start, wave_step=wave_step, wave_count=wave_count)

        # save
        log.info("  Saving to {0:s}...".format(outfile))
        SpectrumFits(spec=resampled).save(outfile)

        # add to CSV
        with open(os.path.join(outdir, 'grid.csv'), 'a') as csv:
            csv.write(filename + ',' + ','.join([str(p) for p in params]) + '\n')





from .filter import Filter
from .fitsspectrum import FitsSpectrum
from .isochrone import Isochrone, MultiIsochrone
from .losvd import LOSVD
from .lsf import LSF, AnalyticalLSF, EmpiricalLSF
from .resultsfits import ResultsFITS
from .spectrum import Spectrum, SpectrumFits, SpectrumFitsHDU, SpectrumBinTableFITS, SpectrumHiResFITS, SpectrumAscii



from .init import Init
from ..component import Component


class InitFromValues(Init):
    """Initializes a component from given values.

    This class, when called, initializes the parameters of a given component to the provided values from a dict.
    """

    def __init__(self, values: dict, *args, **kwargs):
        """Initializes a new Init object.

        Args:
            values: Dictionary of key/value pairs defining initial values for the component's parameters.
        """
        Init.__init__(self, *args, **kwargs)
        self._values = values

    def __call__(self, cmp: Component, filename: str):
        """Initializes parameters of the given component with values from the configuration

        Args:
            cmp: Component to initialize.
            filename: Unused.
        """

        # loop initial values
        for key, val in self._values.items():
            # does parameter exist in component?
            if key in cmp.param_names:
                # set it
                self.log.info('Setting initial value for "%s" of component "%s" to %f...', key, cmp.prefix, val)
                cmp[key] = val


__all__ = ['InitFromValues']



from typing import List, Tuple, Any

from ..object import spexxyObject
from ..grid import GridAxis


class Interpolator(spexxyObject):
    """Base class for all interpolators in spexxy."""
    def __init__(self, cache_level: int = 0, *args, **kwargs):
        """Initializes a new interpolator.

         Args:
            cache_level: Level of caching: 0 means off, 1 means only last dimension, 2 is last 2 dimensions and so on.
                Interpolation might be faster with higher level, but will consume significantly more memory.
        """
        spexxyObject.__init__(self, *args, **kwargs)
        self.cache_level = cache_level
        self.cache = {}

    def axes(self) -> List[GridAxis]:
        """Returns information about the axes.

        Returns:
            List of GridAxis objects describing the grid's axes
        """
        raise NotImplementedError

    def __call__(self, params: Tuple) -> Any:
        """Interpolates at the given parameter set

        Args:
            params: Parameter set to interpolate at

        Returns:
            Interpolation result at given position
        """
        raise NotImplementedError

    def clear_cache(self):
        """Clear cache."""
        self.cache = {}


__all__ = ['Interpolator']



import logging
import numpy as np
import pandas as pd
import argparse

from spexxy.grid import Grid, FilesGrid

log = logging.getLogger(__name__)


def add_parser(subparsers):
    # init parser
    parser = subparsers.add_parser('findholes', help='Find holes in a grid',
                                   formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('grid', type=str, help='Name of grid CSV', default='grid.csv')
    parser.add_argument('output', type=str, help='Output file with holes', default='holes.csv')

    # argparse wrapper for fill_holes
    def run(args):
        grid = FilesGrid(args.grid)
        find_holes(grid, args.output)
    parser.set_defaults(func=run)


def find_holes(grid: Grid, output: str = None) -> pd.DataFrame:
    # loop dimensions
    holes = []
    for axis in range(grid.num_axes()):
        holes.extend(_find_holes_in_axis(grid, axis))

    # sort
    holes = list(sorted(set(holes)))

    # to dataframe
    df = pd.DataFrame(columns=grid.axis_names(), data=holes)

    # store it?
    if output:
        df.to_csv(output, index=False)

    # return it
    return df


def _find_holes_in_axis(grid: Grid, axis: int):
    # get values in this axis
    values = grid.axis_values(axis)

    # get values in all other axes
    other_values = [grid.axis_values(i) for i in range(grid.num_axes()) if i != axis]

    # meshgrid and loop
    holes = []
    mg = np.meshgrid(*other_values)
    for other_params in [tuple(t) for t in zip(*(x.flat for x in mg))]:
        # build list of params
        params = [tuple(list(other_params[:axis]) + [val] + list(other_params[axis:])) for val in values]

        # check for existence
        exists = [p in grid for p in params]

        # find first and last point in grid
        try:
            first = exists.index(True)
            last = len(exists) - exists[-1::-1].index(True) - 1
        except ValueError:
            # True is not in list
            continue

        # finally, find holes, which are all points that do not exists and lie between first and last
        holes.extend([p for i, p in enumerate(params) if first < i < last and not exists[i]])

    # finished
    return list(set(holes))



from .init import Init
from .fromcsv import InitFromCsv
from .frompath import InitFromPath
from .fromvalues import InitFromValues
from .fromvhelio import InitFromVhelio
from .fromyaml import InitFromYAML



from typing import List, Any, Tuple
import pandas as pd
import os
import numpy as np
from astropy.io import fits
from astropy.table import Table

from .grid import Grid, GridAxis
from ..data import Spectrum


class FitsGrid(Grid):
    """Spectrum grid in a single FITS file."""

    def __init__(self, filename: str, norm_to_mean: bool = False, *args, **kwargs):
        """Constructs a new Grid.

        Args:
            filename: Filename of CSV file.
            norm_to_mean: Normalize spectra to their mean.
        """

        # store
        self._norm_to_mean = norm_to_mean

        # expand filename
        filename = os.path.expandvars(filename)

        # load data
        self._spectra, hdr = fits.getdata(filename, header=True)

        # get spec info
        self._wave_start = hdr['CRVAL1']
        self._wave_step = hdr['CDELT1']
        self._wave_mode = Spectrum.Mode.LAMBDA if hdr['CTYPE1'] == 'AWAV' else Spectrum.Mode.LOGLAMBDA

        # load parameters into a pandas table
        self._params = Table(fits.getdata(filename, 'PARAMS')).to_pandas()
        idx_columns = list(self._params.columns.values)

        # add row
        self._params['row'] = range(len(self._params))

        # create axes and init grid
        values = {name: sorted([float(v) for v in self._params[name].unique()]) for name in idx_columns}
        axes = [GridAxis(name=name, values=values[name]) for name in idx_columns]
        Grid.__init__(self, axes, *args, **kwargs)

        # set index
        self._params.set_index(idx_columns, inplace=True)

    @property
    def data(self):
        return self._spectra

    def all(self) -> List[Tuple]:
        """Return all possible parameter combinations.

        Returns:
            All possible parameter combinations.
        """
        return self._params.index.values

    def __contains__(self, params: Tuple) -> bool:
        """Checks, whether the grid contains a given parameter set.

        Args:
            params: Parameter set to check.

        Returns:
            Whether or not the given parameter set exists in the grid.
        """
        return tuple(params) in self._params.index

    def __call__(self, params: Tuple) -> Any:
        """Fetches the value for the given parameter set.

        Args:
            params: Parameter set to catch value for.

        Returns:
            Grid value at given position.
        """

        # get index of parameters
        idx = self._params.loc[tuple(params)].row

        # get flux
        flux = self._spectra[idx, :]

        # create spectrum
        spec = Spectrum(flux=flux, wave_start=self._wave_start, wave_step=self._wave_step, wave_mode=self._wave_mode)

        # normalize?
        if self._norm_to_mean:
            spec.norm_to_mean()

        # return it
        return spec


__all__ = ['FitsGrid']
